import unittest
from unittest.mock import MagicMock
from datetime import timedelta

from osgar.bus import Bus

from subt.scan_mixer import ScanMixer


class ScanMixerTest(unittest.TestCase):

    def test_usage(self):
        bus = Bus(MagicMock(write=MagicMock(return_value=timedelta())))
        tester = bus.handle('tester')
        tester.register("scan", "rs_scan")
        mixer = ScanMixer(config={}, bus=bus.handle('mixer'))
        bus.connect("tester.scan", "mixer.scan")
        bus.connect("tester.rs_scan", "mixer.rs_scan")
        bus.connect("mixer.scan", "tester.mixed")
        mixer.start()
        scan = [10, 10, 10, 10]
        tester.publish("scan", scan)
        _, channel, data = tester.listen()
        self.assertEqual(data, scan)
        mixer.request_stop()
        mixer.join()

    def test_mixing(self):
        log = MagicMock()
        log.write.return_value = timedelta()
        bus = Bus(log)
        tester = bus.handle('tester')
        tester.register("scan", "rs_scan")
        mixer = ScanMixer(config={}, bus=bus.handle('mixer'))
        bus.connect("tester.scan", "mixer.scan")
        bus.connect("tester.rs_scan", "mixer.rs_scan")
        bus.connect("mixer.scan", "tester.mixed")
        mixer.start()
        scan = 4*[10]
        rs_scan = 3*[8]
        tester.publish("rs_scan", rs_scan)
        tester.publish("scan", scan)
        _, channel, data = tester.listen()
        self.assertEqual(data, 4*[8])
        mixer.request_stop()
        mixer.join()

    def test_infinity(self):
        log = MagicMock()
        log.write.return_value = timedelta()
        bus = Bus(log)
        tester = bus.handle('tester')
        tester.register("scan", "rs_scan")
        mixer = ScanMixer(config={}, bus=bus.handle('mixer'))
        bus.connect("tester.scan", "mixer.scan")
        bus.connect("tester.rs_scan", "mixer.rs_scan")
        bus.connect("mixer.scan", "tester.mixed")

        mixer.start()
        scan = 3*[10]+[0]
        rs_scan = 3*[8]
        tester.publish("rs_scan", rs_scan)
        tester.publish("scan", scan)
        _, channel, data = tester.listen()
        self.assertEqual(data, 4*[8])

        scan = 4*[10]
        rs_scan = 2*[8]+[0]
        tester.publish("rs_scan", rs_scan)
        tester.publish("scan", scan)
        _, channel, data = tester.listen()
        self.assertEqual(data, 3*[8]+[10])

        mixer.request_stop()
        mixer.join()

    def test_old_rs_scan(self):
        log = MagicMock()
        log.write.return_value = timedelta()
        bus = Bus(log)
        tester = bus.handle('tester')
        tester.register("scan", "rs_scan")
        mixer = ScanMixer(config={}, bus=bus.handle('mixer'))
        bus.connect("tester.scan", "mixer.scan")
        bus.connect("tester.rs_scan", "mixer.rs_scan")
        bus.connect("mixer.scan", "tester.mixed")

        mixer.start()
        scan = 3*[10]+[0]
        rs_scan = 3*[8]
        tester.publish("rs_scan", rs_scan)
        log.write.return_value = timedelta(seconds=1)
        tester.publish("scan", scan)
        _, channel, data = tester.listen()
        self.assertEqual(data, 3*[10]+[0])

        mixer.request_stop()
        mixer.join()

# vim: expandtab sw=4 ts=4



"""
  GPS Driver
"""

from threading import Thread
import struct

from osgar.bus import BusShutdownException


INVALID_COORDINATES = [None, None]
BIN_PREAMBULE = bytes([0xB5, 0x62])


def checksum(s):
    sum = 0
    for ch in s:
        sum ^= ch
    return b"%02X" % (sum)


def str2ms(s):
    'convert DDMM.MMMMMM string to arc milliseconds(int)'
    if s == b'':  # unknown position
        return None
    dm, frac = (b'0000' + s).split(b'.')
    try:
        return round((int(dm[:-2]) * 60 + float(dm[-2:] + b'.' + frac)) * 60000)
    except Exception as e:
        print(e)
        return None


def parse_line(line):
    assert line.startswith(b'$GNGGA') or line.startswith(b'$GPGGA'), line
    assert checksum(line[1:-3]) == line[-2:], (line, checksum(line[1:-3]))
    s = line.split(b',')
    coord = [str2ms(s[4]), str2ms(s[2])]
    return coord


def parse_bin(data):
    assert data.startswith(BIN_PREAMBULE), data
    c, i, size = struct.unpack_from('<BBH', data, 2)
    assert len(data) == size + 8, (len(data), size + 8)
    assert c == 1, c  # class = 1 ... NAVigation messages
    assert i in [3, 0x30, 6, 7, 0x34, 0x35, 1, 2, 0x13, 0x14, 4, 0x11, 0x12, 0x20, 0x23, 0x24, 0x21,
                0x26, 0x22, 0x9, 0x3B, 0x3C, 0x39, 0x61, ], hex(i)  # ID
#    print(hex(i))

    # TODO verify checksum!
    payload = data[6:-2]

    # 31.18.20 UBX-NAV-STATUS (0x01 0x03)
    # 31.18.20.1 Receiver Navigation Status
    # Receiver Navigation Status
    if i == 0x03:
        fix = payload[4]
        #assert fix in [2, 3], fix  # 2D, 3D
        if fix not in [2, 3]:
            print("GPS no fix!")

    # 31.18.21 UBX-NAV-SVINFO (0x01 0x30)
    # 31.18.21.1 Space Vehicle Information
    # Information about satellites used or visible
    if i == 0x30:
        return None

    # 31.18.15 UBX-NAV-RELPOSNED (0x01 0x3C)
    # 31.18.15.1 Relative Positioning Information in NED frame
    if i == 0x3C:
        assert len(payload) == 40, len(payload)
        assert payload[0] == 0, payload[0]  # version
        iTOW, rel_pos_north_cm, rel_pos_east_cm = struct.unpack_from('<Iii', payload, 4)
        accN, accE = struct.unpack_from('<II', payload, 24)
        return {'rel_position': [rel_pos_east_cm, rel_pos_north_cm]}

    # 31.18.30 UBX-NAV-VELNED (0x01 0x12)
    # 31.18.30.1 Velocity Solution in NED
    if i == 0x12:
        assert len(payload) == 36, len(payload)
        iTOW, velN, velE, velD = struct.unpack_from('<Iiii', payload, 0)
        gSpeed, heading, sAcc, cAcc = struct.unpack_from('<IiII', payload, 20)
        #print(gSpeed, heading/1e5, sAcc, cAcc/1e5)
        return None  # do not integrate for now

    # 31.18.14 UBX-NAV-PVT (0x01 0x07)
    # 31.18.14.1 Navigation Position Velocity Time Solution
    if i == 0x07:
        assert len(payload) == 92, len(payload)
        gSpeed, heading, sAcc, cAcc = struct.unpack_from('<IiII', payload, 60)
        #print('xx', gSpeed, heading/1e5, sAcc, cAcc/1e5)
        return None  # do not integrate for now


def split_buffer(data):
    # in dGPS there is a block of binary data so stronger selection is required
    start_nmea = max(data.find(b'$GNGGA'), data.find(b'$GPGGA'))
    start_bin = data.find(BIN_PREAMBULE)
    if start_nmea < 0 or (0 <= start_bin < start_nmea):
        if start_bin < 0 or start_bin + 8 >= len(data):
            return data, b''
        else:
            # extract binary data: preambule, class, ID, len, payload, checksum
            c, i, size = struct.unpack_from('<BBH', data, start_bin + 2)
            end = start_bin + 6 + size + 2
            if end >= len(data):
                return data, b''
            return data[end:], data[start_bin:end]

    start = start_nmea
    end = data[start:-2].find(b'*')
    if end < 0:
        return data, b''
    return data[start+end+3:], data[start:start+end+3]


class GPS(Thread):
    def __init__(self, config, bus):
        bus.register('position')
        Thread.__init__(self)
        self.setDaemon(True)

        self.bus = bus
        self.buf = b''

    def process_packet(self, line):
        if line.startswith(b'$GNGGA') or line.startswith(b'$GPGGA'):
            coords = parse_line(line)
            return {'position': coords}
        elif line.startswith(BIN_PREAMBULE):
            return parse_bin(line)
        return None

    def process_gen(self, data):
        self.buf, packet = split_buffer(self.buf + data)
        while len(packet) > 0:
            ret = self.process_packet(packet)
            if ret is not None:
                for k, v in ret.items():
                    yield k, v
            self.buf, packet = split_buffer(self.buf)  # i.e. process only existing buffer now

    def run(self):
        try:
            while True:
                packet = self.bus.listen()  # there should be some timeout and in case of failure send None
                dt, __, data = packet
                for name, out in self.process_gen(data):
                    assert out is not None
                    self.bus.publish(name, out)
        except BusShutdownException:
            pass

    def request_stop(self):
        self.bus.shutdown()


def print_output(packet):
    print(packet)

# vim: expandtab sw=4 ts=4



"""
  OSGAR Drone for control of flight height
"""
# copy of drone_height.py
import math

from osgar.node import Node

HEIGHT = 2.0
EMERGENCY_HEIGHT = 0.5
MAX_ANGULAR = 0.7
MAX_VERTICAL = 1
PID_P = 1.5  # 2.0  # 0.5


def altitude_from_pressure(p):
    # Model and parameters taken from Wikipedia:
    # https://en.wikipedia.org/wiki/Barometric_formula
    # I have not yet found corresponding atmospheric model and constants
    # in Ignition Gazebo.

    # Reference pressure. [Pa]
    Pb = 101325.00
    # Reference temperature. [K]
    Tb = 288.15
    # Gravitational acceleration. [m/s^2]
    g0 = 9.80665
    # Universal gas constant. [J/(mol*K)]
    R = 8.3144598
    # Molar mass of air. [kg/mol]
    M = 0.0289644
    # Height of reference level. [m]
    hb = 0

    # The pressure model is:
    #   p = Pb * exp(-g0 * M * (h - hb) / (R * Tb))
    # Therefore:
    h = (math.log(p) - math.log(Pb)) * (R * Tb) / (-g0 * M) + hb

    return h


class Drone(Node):
    #class that listens all topics needed for controlling drone height
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register("desired_speed_3d")

        self.count = 0 # count of bottom messages
        self.lastScanDown = 0.0
        self.lastScanUp = 0.0
        self.started = False

        self.debug_arr = []
        self.debug_arr2 = []
        self.verbose = False
        self.z = None  # last Z coordinate
        self.altitude = None  # based on air_pressure
        self.desired_z_speed = None  # i.e. keep safe distance from up and down obstacles

        # keep last command for independent update of XY and Z
        self.linear = [0.0, 0.0, 0.0]
        self.angular = [0.0, 0.0, 0.0]

    def send_speed_cmd(self):
        if self.started:
            H, E, A = 0, 0, 0
            desiredVel = None
            if self.desired_z_speed is None:
                height = min(HEIGHT, (self.lastScanDown + self.lastScanUp) / 2)
                diff = height - self.lastScanDown
                H = diff
                desiredVel = PID_P * diff
            else:
                if min(self.lastScanDown, self.lastScanUp) < EMERGENCY_HEIGHT:
                    print(self.time, 'Emergency limit', self.lastScanDown, self.lastScanUp)
                    # temporary switch to keep safe height from bottom/top
                    down = self.lastScanDown - EMERGENCY_HEIGHT
                    up = self.lastScanUp - EMERGENCY_HEIGHT
                    if up < 0 and down < 0:
                        diff = (up - down) / 2
                    elif up < 0:
                        diff = up
                    else:
                        diff = -down
                    E = diff
                    desiredVel = PID_P * diff
                else:
                    A = self.desired_z_speed
                    desiredVel = self.desired_z_speed

            desiredVel = max(-MAX_VERTICAL, min(MAX_VERTICAL, desiredVel))
            self.linear[2] = desiredVel

            #this is because drone can't handle too aggressive rotation
            if self.angular[2] > MAX_ANGULAR:
                self.angular[2] = MAX_ANGULAR
            elif self.angular[2] < -MAX_ANGULAR:
                self.angular[2] = -MAX_ANGULAR

            self.publish('desired_speed_3d', [self.linear, self.angular])
            if self.verbose:
                self.debug_arr2.append((self.count, H, E, A))

    def on_bottom_scan(self, scan):
        self.count += 1
        self.lastScanDown = scan[0]
        self.send_speed_cmd()

    def on_top_scan(self, scan):
        self.lastScanUp = scan[0]

    def on_desired_speed(self, data):
        self.linear = [data[0]/1000, 0.0, 0.0]
        self.angular = [0.0, 0.0, math.radians(data[1]/100.0)]

        if self.verbose:
            self.debug_arr.append((self.time, self.altitude, self.lastScanDown, self.lastScanUp, self.z))

        if data != [0, 0]:
            self.started = True

        self.send_speed_cmd()

    def on_desired_z_speed(self, z_speed):
        self.desired_z_speed = z_speed

    def on_pose3d(self, data):
        xyz, quat = data
        self.z = xyz[2]

    def on_air_pressure(self, data):
        self.altitude = altitude_from_pressure(data)

    def on_desired_altitude(self, data):
        pass  # self.desired_altitude already updated by Node

    def update(self):
        channel = super().update()

        handler = getattr(self, "on_" + channel, None)
        if handler is not None:
            handler(getattr(self, channel))
        else:
            assert False, "unknown input {channel}"


    def draw(self):
        import matplotlib.pyplot as plt

        #t = [a[0].total_seconds() for a in self.debug_arr]
        #height = [(a[1], a[1] - a[2], a[1] + a[3], a[4]) for a in self.debug_arr]
        #line_obj = plt.plot(t, height, '-o', linewidth=2)
        #plt.legend(iter(line_obj), ('altitude', 'alt - bottom', 'alt + top', 'pose3D.z'))
        #plt.xlabel('time (s)')
        x = [count/50.0 for count, H, E, A in self.debug_arr2]  # bottom lidar runnig at 50Hz
        a = [A for count, H, E, A in self.debug_arr2]
        e = [E for count, H, E, A in self.debug_arr2]
        h = [H for count, H, E, A in self.debug_arr2]
        plt.plot(x, a)
        plt.plot(x, e)
        plt.plot(x, h)
        plt.xlabel('sim time (s)')
        plt.legend()
        plt.show()

# vim: expandtab sw=4 ts=4



"""
  Driver for a small tank Maria

  Text Communication Protocol
  Version 0 uses ASCII protocol between the host computer (JetsonNano) and Arduino motor controller.
  There is no speed control implemented on Arduino at the moment, so raw PWM is used:
     "<PWM left> <PWM right> <desired LED status>\n"
  The value for PWM is in range -255..255
  LED status: 0=OFF, 1=green, 2=red.

  The board regularly responds (20Hz?) with status line:
      "<timestamp> <encoder left> <encoder right> <buttons mask>\n"
  The timestamp is 32bit unsigned int in milliseconds.
  The encoders values are 32bit signed int and accumulative.
  The buttons mask is currently not used (first remote tests are with Emergency STOP pressed).

  The plan is to abandon this text protocol and use binary structures inspired by ATmega8board.
"""

import math

from osgar.node import Node
from osgar.bus import BusShutdownException


ENC_SCALE = 0.00052
WHEEL_DISTANCE = 0.4  # 50cm long, 40cm wide
UPDATE_RATE = 20.0  # Hz
MIN_PWM = 10
MAX_PWM = 60

GREEN_LED = 1
RED_LED = 2


class SpeedControl:
    def __init__(self):
        self.min_pwm = MIN_PWM
        self.max_pwm = MAX_PWM
        self.step_pwm = 2
        self.tolerance = 5  # do not change PWM output if it is in +/- tolerance

        self.desired_ticks = 0
        self.last_pwm = 0

    def set_desired_ticks(self, ticks):
        """desired encoder ticks per fixed period"""
        self.desired_ticks = ticks

    def update(self, encoder_ticks):
        """Regular status of encoders ticks, return new PWM"""
        pwm = self.last_pwm
        if self.desired_ticks == 0:
            pwm = 0
        elif self.desired_ticks > 0:
            if encoder_ticks < -self.tolerance:
                # change of polarity, transit via 0
                pwm = 0
            elif abs(encoder_ticks - self.desired_ticks) > self.tolerance:
                if encoder_ticks > self.desired_ticks:
                    pwm = max(self.last_pwm - self.step_pwm, self.min_pwm)
                else:
                    pwm = max(self.last_pwm + self.step_pwm, self.min_pwm)
        else:  # self.desired_ticks < 0
            if encoder_ticks > self.tolerance:
                pwm = 0
            elif abs(encoder_ticks - self.desired_ticks) > self.tolerance:
                if abs(encoder_ticks) > abs(self.desired_ticks):
                    pwm = -max(-self.last_pwm - self.step_pwm, self.min_pwm)
                else:
                    pwm = -max(-self.last_pwm + self.step_pwm, self.min_pwm)
        pwm = max(-self.max_pwm, min(self.max_pwm, pwm))  # clip limits
        self.last_pwm = pwm
        return pwm


class RobotMaria(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register('pose2d', 'emergency_stop', 'encoders', 'raw')

        # commands
        self.desired_speed = 0.0  # m/s
        self.desired_angular_speed = 0.0
        self.sc_left = SpeedControl()
        self.sc_right = SpeedControl()

        # status
        self.emergency_stop = None  # uknown state
        self.pose2d = (0.0, 0.0, 0.0)  # x, y in meters, heading in radians (not corrected to 2PI)
        self.buttons = None

        self.verbose = False  # should be in Node
        self.buf = b''
        self.encoder = None
        self.led = GREEN_LED
        self.milisec = None

    def send_pose2d(self):
        x, y, heading = self.pose2d
        self.publish('pose2d', [round(x*1000), round(y*1000),
                                round(math.degrees(heading)*100)])


    def compute_pose2d(self, left, right):
        """Update internal pose2d"""
        if left is None or right is None:
            return False, None, None
        x, y, heading = self.pose2d

        metricL = ENC_SCALE * left
        metricR = ENC_SCALE * right

        dist = (metricL + metricR)/2.0
        angle = (metricR - metricL)/WHEEL_DISTANCE

        # advance robot by given distance and angle
        if abs(angle) < 0.0000001:  # EPS
            # Straight movement - a special case
            x += dist * math.cos(heading)
            y += dist * math.sin(heading)
            #Not needed: heading += angle
        else:
            # Arc
            r = dist / angle
            x += -r * math.sin(heading) + r * math.sin(heading + angle)
            y += +r * math.cos(heading) - r * math.cos(heading + angle)
            heading += angle  # intentionally not normalized
            #  Explanation:
            #    Pose2D from odometry is one exception where you can collect
            #  history, i.e. that robot turned twice 360 degrees for example.
            #  This is different to compass where you have only current heading.
        return True, (x, y, heading), (dist, angle)

    def update_pwm0(self, encoders_left, encoders_right):
        STEP = int(min(255, abs(100 * self.desired_speed)))  # test PWM
        if STEP == 0:
            STEP = 200  # hardcoded for rotaion in place
        if self.desired_speed > 0:
            left, right = STEP, STEP
        elif self.desired_speed < 0:
            left, right = -STEP, -STEP
        elif self.desired_angular_speed > 0:
            left, right = -STEP, STEP
        elif self.desired_angular_speed < 0:
            left, right = STEP, -STEP
        else:
            left, right = 0, 0  # STOP

        self.publish('raw', b'%d %d %d\n' % (left, right, self.led))

    def update_pwm(self, encoders_left, encoders_right):
        left = self.sc_left.update(encoders_left)
        right = self.sc_right.update(encoders_right)
        self.publish('raw', b'%d %d %d\n' % (left, right, self.led))

    def get_packet(self, buf):
        s = buf.split(b'\n')
        if len(s) < 2:
            return None, buf
        return s[0].strip(), b'\n'.join(s[1:])

    def slot_desired_speed(self, timestamp, data):
        self.desired_speed, self.desired_angular_speed = data[0]/1000.0, math.radians(data[1]/100.0)
        left = self.desired_speed - self.desired_angular_speed * WHEEL_DISTANCE/2
        right = self.desired_speed + self.desired_angular_speed * WHEEL_DISTANCE/2
        self.sc_left.set_desired_ticks(left/ENC_SCALE/UPDATE_RATE)
        self.sc_right.set_desired_ticks(right/ENC_SCALE/UPDATE_RATE)

    def slot_raw(self, timestamp, data):
        self.buf += data
        packet, self.buf = self.get_packet(self.buf)
        while packet is not None:
            try:
                d = [int(x) for x in packet.split()]
            except ValueError:
                print('Invalid packet', packet.split())
                packet, self.buf = self.get_packet(self.buf)
                continue

            if self.verbose:
                print(d)
            assert len(d) == 4, d
            prev = self.encoder
            self.milisec = d[0]
            self.encoder = d[1], d[2]
            if prev is not None:
                left, right = self.encoder[0] - prev[0], self.encoder[1] - prev[1]
                self.publish('encoders', [left, right])
                valid, pose2d, dist_angle = self.compute_pose2d(left, right)
                if valid:
                    self.pose2d = pose2d
                self.send_pose2d()
                self.update_pwm(left, right)

            packet, self.buf = self.get_packet(self.buf)

    def run(self):
        try:
            while True:
                self.time, channel, data = self.listen()

                if channel == 'raw':
                    self.slot_raw(self.time, data)
                elif channel == 'desired_speed':
                    self.slot_desired_speed(self.time, data)
                else:
                    assert False, channel  # unsupported channel
        except BusShutdownException:
            pass

# vim: expandtab sw=4 ts=4



"""
  Timer
"""
from threading import Thread

from osgar.bus import BusShutdownException


class MyTimer(Thread):
    def __init__(self, config, bus):
        Thread.__init__(self)
        self.setDaemon(True)
        self.bus = bus
        self.sleep_time = config['sleep']

    def run(self):
        try:
            while self.bus.is_alive():
                self.bus.publish('tick', None)
                self.bus.sleep(self.sleep_time)

        except BusShutdownException:
            pass

    def request_stop(self):
        self.bus.shutdown()

# vim: expandtab sw=4 ts=4




"""
  Follow Me (primary target Eduro robot)
"""
import math
from datetime import timedelta

from osgar.node import Node

# Notes:
# Expects SICK laser scan with 270 degrees field of view mounted in front.


# TODO shared place for multiple applications
class EmergencyStopException(Exception):
    pass


def min_dist(laser_data):
    if len(laser_data) > 0:
        # remove ultra near reflections and unlimited values == 0
        laser_data = [x if x > 10 else 10000 for x in laser_data]
        return min(laser_data)
    return 0


class FollowMe(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register('desired_speed')
        self.last_position = [0, 0, 0]  # proper should be None, but we really start from zero
        self.raise_exception_on_stop = False
        self.verbose = False
        self.last_scan = None

    def update(self):
        channel = super().update()  # define self.time
        if channel == 'pose2d':
            self.last_position = self.pose2d
        elif channel == 'scan':
            if self.verbose:
                print(min_dist(self.scan)/1000.0)
            self.last_scan = self.scan
        elif channel == 'emergency_stop':
            if self.raise_exception_on_stop and self.emergency_stop:
                raise EmergencyStopException()

    def wait(self, dt):  # TODO refactor to some common class
        if self.time is None:
            self.update()
        start_time = self.time
        while self.time - start_time < dt:
            self.update()

    def send_speed_cmd(self, speed, angular_speed):
        return self.bus.publish('desired_speed', 
                [round(speed*1000), round(math.degrees(angular_speed)*100)])

    def followme(self):
        print("Follow Me!")

        SCAN_SIZE = 811  # TODO config
        SCANS_PER_DEG = SCAN_SIZE//270

        # limit tracking to front 180deg only due to mounting (back laser is blocked by robot body)
        LIMIT_LOW = 0  # SCAN_SIZE//6
        LIMIT_HIGH = SCAN_SIZE  # 5*SCAN_SIZE//6
        CLOSE_REFLECTIONS = 10  # ignore readings closer than 10mm, where 0 = infinite (no response)

        thresholds = []
        for i in range(SCAN_SIZE):
            if LIMIT_LOW <= i <= LIMIT_HIGH:
                deg = -135 + 270 * i / SCAN_SIZE
                rad = math.radians(deg)
                thresh = 1000 * (0.17 + 0.17 * max(0, math.cos(rad)))  # [mm]
            else:
                thresh = 0
            thresholds.append(thresh)

        masterAngleOffset = 0  # TODO
        index = None
        while True:
            if self.last_scan is not None:
                assert len(self.last_scan) == SCAN_SIZE, len(self.last_scan)
                near = 10.0
                if index is None:
                    low = LIMIT_LOW
                    high = LIMIT_HIGH
                else:
                    # search in +/- 10deg sector
                    low = max(LIMIT_LOW, index - 10 * SCANS_PER_DEG)
                    high = min(LIMIT_HIGH, index + 10 * SCANS_PER_DEG)

                for i in range(low, high):
                    x = self.last_scan[i]
                    if CLOSE_REFLECTIONS < x < near * 1000:
                        near = x / 1000.0
                        index = i
                maxnear = min( (x for x in self.last_scan if x > CLOSE_REFLECTIONS) ) / 1000.0

                if self.verbose:
                    print(near, maxnear, index)

                if near > 1.3 or any(x < thresh for (x, thresh) in zip(self.last_scan, thresholds) if x > CLOSE_REFLECTIONS):
                    self.send_speed_cmd(0.0, 0.0)
                else:
                    angle = math.radians(270 * index/SCAN_SIZE - 135) + masterAngleOffset
                    desiredAngle = 0
                    desiredDistance = 0.4
                    speed = 0.2 + 2 * (near - desiredDistance)
                    rot = 1.5 * (angle - desiredAngle)
                    if speed < 0:
                        speed = 0
                    self.send_speed_cmd(speed, rot)

                self.last_scan = None
            self.update()

    def run(self):
        try:
            self.raise_exception_on_stop = True
            self.followme()
        except EmergencyStopException:
            print('!!!Emergency STOP!!!')
            self.raise_exception_on_stop = False
            self.send_speed_cmd(0.0, 0.0)
            self.wait(timedelta(seconds=1))


if __name__ == "__main__":
    from osgar.launcher import launch

    launch(app=FollowMe, description='Follow Me', prefix='followme-')

# vim: expandtab sw=4 ts=4



import unittest
from unittest.mock import MagicMock

from subt.breadcrumbs import Breadcrumbs
from osgar.lib import quaternion

class BreadcrumbsTest(unittest.TestCase):

    def test_deploy(self):
        bus = bus=MagicMock()
        bread = Breadcrumbs(bus=bus, config={'step_sec':10})
        bread.on_sim_time_sec(3)
        bus.publish.assert_not_called()
        bread.on_sim_time_sec(14)
        bus.publish.assert_called()
        bus.publish.reset_mock()
        bread.on_sim_time_sec(15)
        bus.publish.assert_not_called()

    def test_breadcrumbs_dist(self):
        bus = bus=MagicMock()
        bread = Breadcrumbs(bus=bus, config={'radius':10})
        self.assertEqual(bread.locations, [[0, 0, 0]])

        bread.on_pose3d([[1, 0, 0], quaternion.identity()])
        self.assertEqual(bread.locations, [[0, 0, 0]])
        bus.publish.assert_not_called()

        bread.on_pose3d([[11, 0, 0], quaternion.identity()])
        self.assertEqual(bread.locations, [[0, 0, 0], [11, 0, 0]])
        bus.publish.assert_called()

    def test_external_breadcrumbs(self):
        bus = bus=MagicMock()
        bread = Breadcrumbs(bus=bus, config={'radius':10})
        self.assertEqual(bread.locations, [[0, 0, 0]])

        bus.publish.reset_mock()
        bread.on_external([11, 0, 0])
        self.assertEqual(bread.locations, [[0, 0, 0], [11, 0, 0]])

        bread.on_external([5, 0, 0])  # i.e. not within the 10m radius, but it is already placed
        self.assertEqual(bread.locations, [[0, 0, 0], [11, 0, 0], [5, 0, 0]])

    def test_limited_size(self):
        bus = MagicMock()
        bread = Breadcrumbs(bus=bus, config={'radius':10})
        self.assertEqual(bread.locations, [[0, 0, 0]])

        for x in range(15, 1000, 15):
            bread.on_pose3d([[x, 0, 0], quaternion.identity()])

        self.assertEqual(len(bread.locations), 1 + 6)

    def test_num_available(self):
        bus = MagicMock()
        bread = Breadcrumbs(bus=bus, config={'radius':10})
        self.assertEqual(bread.num_avail, 6)

        bread = Breadcrumbs(bus=bus, config={'radius':10, 'num': 12})
        self.assertEqual(bread.num_avail, 12)


# vim: expandtab sw=4 ts=4



from osgar.logger import LogReader, lookup_stream_id
from osgar.lib.serialize import deserialize

CELL_PREFIX = "PhoneArtifact"

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description='Parse logged WiFi AP scans and extract "Cell Phone" artifact')
    parser.add_argument('logfile', help='recorded log file')
    args = parser.parse_args()

    only_stream = lookup_stream_id(args.logfile, 'wifi.wifiscan')
    best = None
    with LogReader(args.logfile, only_stream_id=only_stream) as log:
        for timestamp, stream_id, data in log:
            wifi_list = deserialize(data)
            for name, signal in wifi_list:
                if name.startswith(CELL_PREFIX):
                    print(timestamp, name, signal)
                    if best is None or signal > best[2]:
                        best = timestamp, name, signal
    print('--------------------')
    if best is not None:
        print('best:', best[0], best[1], best[2])
    else:
        print("No Cell Phone, sorry")

# vim: expandtab sw=4 ts=4




import datetime
import math
import unittest
from unittest.mock import MagicMock, patch, call

from osgar.bus import Bus
from subt.drone import Drone, MAX_ANGULAR, altitude_from_pressure


class DroneTest(unittest.TestCase):

    def test_usage(self):
        logger = MagicMock()
        bus = Bus(logger)
        logger.write = MagicMock(return_value=datetime.timedelta(microseconds=9721))
        c = Drone(bus=bus.handle('drone'), config={})
        tester = bus.handle('tester')
        tester.register('desired_speed')
        bus.connect('tester.desired_speed', 'drone.desired_speed')
        bus.connect('drone.desired_speed_3d', 'tester.desired_speed_3d')
        c.start()
        tester.publish('desired_speed', [1000, 9000])
        c.request_stop()
        c.join()
        self.assertEqual(tester.listen()[2], [[1.0, 0.0, 0.0], [0.0, 0.0, MAX_ANGULAR]])

    def test_altitude_from_pressure(self):
        self.assertEqual(altitude_from_pressure(101323.49836826918), 0.12500215269779752)

    def test_regular_update(self):
        bus = MagicMock()
        drone = Drone(bus=bus, config={})
        drone.on_desired_speed([1000, 0])  # fly forward for ever
        bus.publish.assert_called_with('desired_speed_3d', [[1.0, 0.0, 0.0], [0.0, 0.0, 0.0]])

        bus.reset_mock()
        drone.on_top_scan([10.1])  # order is unfortunately not defined -> extra trigger is on bottom scan
        drone.on_bottom_scan([0.1])
        bus.publish.assert_called_with('desired_speed_3d', [[1.0, 0.0, 1.0], [0.0, 0.0, 0.0]])


# vim: expandtab sw=4 ts=4



import unittest
from unittest.mock import patch, MagicMock
import datetime
from collections import namedtuple

import numpy as np
import cv2

from osgar.drivers.realsense import RealSense
from osgar.bus import Bus


Pose = namedtuple('Pose', [
    'acceleration',
    'angular_acceleration',
    'angular_velocity',
    'mapper_confidence',
    'rotation',
    'tracker_confidence',
    'translation',
    'velocity',
    ])
Rotation = namedtuple('Rotation', ['x', 'y', 'z', 'w'])
Translation = namedtuple('Translation', ['x', 'y','z'])
Velocity = namedtuple('Velocity', ['x', 'y','z'])
AngularVelocity = namedtuple('AngularVelocity', ['x', 'y','z'])
Acceleration = namedtuple('Acceleration', ['x', 'y','z'])
AngularAcceleration = namedtuple('AngularAcceleration', ['x', 'y','z'])


class RealSenseTest(unittest.TestCase):
    def test_detect_pose(self):
        bus = MagicMock()
        with patch('osgar.drivers.realsense.rs') as rs:
            ctx = rs.context.return_value
            device = MagicMock()
            device.get_info.return_value = "T200"
            ctx.query_devices.return_value = [device]
            c = RealSense(bus=bus.handle('rs'), config={})
            c.start()
            self.assertTrue(c.pose_pipeline.start.called)
            self.assertEqual(c.depth_pipeline, None)
            c.request_stop()
            c.join()

    def test_detect_depth(self):
        bus = MagicMock()
        with patch('osgar.drivers.realsense.rs') as rs:
            ctx = rs.context.return_value
            device = MagicMock()
            device.get_info.return_value = "D400"
            ctx.query_devices.return_value = [device]
            c = RealSense(bus=bus.handle('rs'), config={})
            c.start()
            self.assertTrue(c.depth_pipeline.start.called)
            self.assertEqual(c.pose_pipeline, None)
            c.request_stop()
            c.join()

    def test_detect_pose_depth(self):
        bus = MagicMock()
        with patch('osgar.drivers.realsense.rs') as rs:
            ctx = rs.context.return_value
            pose_device = MagicMock()
            pose_device.get_info.return_value = "T200"
            depth_device = MagicMock()
            depth_device.get_info.return_value = "D400"
            ctx.query_devices.return_value = [pose_device, depth_device]
            c = RealSense(bus=bus.handle('rs'), config={})
            c.start()
            self.assertTrue(c.pose_pipeline.start.called)
            self.assertTrue(c.depth_pipeline.start.called)
            c.request_stop()
            c.join()

    def test_pose2d_moves(self):
        logger = MagicMock()
        logger.write = MagicMock(return_value=datetime.timedelta(microseconds=9721))
        bus = Bus(logger)
        moves = [
            [Translation(0, 0, -1), [1000, 0, 0]], # forward
            [Translation(-1, 0, 0), [0, 1000, 0]], # left
        ]
        c = RealSense(bus=bus.handle('rs'), config={})
        tester = bus.handle('tester')
        bus.connect('rs.pose2d', 'tester.pose2d')

        for input, output in moves:
            frame = MagicMock()
            frame.get_frame_number.return_value = 1
            frame.get_timestamp.return_value = 1
            pose_frame = frame.as_pose_frame.return_value
            pose_frame.get_pose_data.return_value = Pose(
                Acceleration(0, 0, 0),
                AngularAcceleration(0, 0, 0),
                AngularVelocity(0, 0, 0),
                0,
                Rotation(0, 0, 0, 1),
                0,
                input,
                Velocity(0, 0, 0))
            frame.get_timestamp.return_value = 0
            frame.get_frame_number.return_value = 0
            c.pose_callback(frame)
            dt, channel, pose2d = tester.listen()
            self.assertEqual(channel, 'pose2d')
            self.assertEqual(pose2d, output)

    def test_orientation(self):
        logger = MagicMock()
        logger.write = MagicMock(return_value=datetime.timedelta(microseconds=9721))
        bus = Bus(logger)
        moves = [
            # heading zero
            [Rotation(0, 0, 0, 1), [0, 0, 0, 1]],

            # looking left 90 degrees
            # t265 turns around y axis +90
            # osgar turns round z axis +90
            [Rotation(0, 0.7071068, 0, 0.7071068), [0, 0, 0.7071068, 0.7071068]],

            # looking up 90 degrees
            # t265 turns around x axis +90
            # osgar turns round y axis -90
            [Rotation(0.7071068, 0, 0, 0.7071068), [0, -0.7071068, 0, 0.7071068]]
        ]
        c = RealSense(bus=bus.handle('rs'), config={})
        tester = bus.handle('tester')
        bus.connect('rs.orientation', 'tester.orientation')
        for input, output in moves:
            frame = MagicMock()
            pose_frame = frame.as_pose_frame.return_value
            pose_frame.get_pose_data.return_value = Pose(
                Acceleration(0, 0, 0),
                AngularAcceleration(0, 0, 0),
                AngularVelocity(0, 0, 0),
                0,
                input,
                0,
                Translation(0, 0, 0),
                Velocity(0, 0, 0))
            frame.get_timestamp.return_value = 0
            frame.get_frame_number.return_value = 0
            c.pose_callback(frame)
            dt, channel, orientation = tester.listen()
            self.assertEqual(channel, 'orientation')
            self.assertEqual(orientation, output)

    def test_pose2d_heading(self):
        logger = MagicMock()
        logger.write = MagicMock(return_value=datetime.timedelta(microseconds=9721))
        bus = Bus(logger)
        moves = [
            [Rotation(0, 0, 0, 1), [0, 0, 0]], # heading zero
            [Rotation(0, 0.7071068, 0, 0.7071068), [0, 0, 90*100]],   # facing left
            [Rotation(0, -0.7071068, 0, 0.7071068), [0, 0, -90*100]], # facing right
            [Rotation(0, -0.9999619, 0, 0.0087265), [0, 0, -179*100]],  # facing backwards
        ]
        c = RealSense(bus=bus.handle('rs'), config={})
        tester = bus.handle('tester')
        bus.connect('rs.pose2d', 'tester.pose2d')
        for input, output in moves:
            frame = MagicMock()
            pose_frame = frame.as_pose_frame.return_value
            pose_frame.get_pose_data.return_value = Pose(
                Acceleration(0, 0, 0),
                AngularAcceleration(0, 0, 0),
                AngularVelocity(0, 0, 0),
                0,
                input,
                0,
                Translation(0, 0, 0),
                Velocity(0, 0, 0))
            frame.get_timestamp.return_value = 0
            frame.get_frame_number.return_value = 0
            c.pose_callback(frame)
            dt, channel, pose2d = tester.listen()
            self.assertEqual(channel, 'pose2d')
            self.assertEqual(pose2d, output)

    def test_pose3d(self):
        logger = MagicMock()
        logger.write = MagicMock(return_value=datetime.timedelta(microseconds=9721))
        bus = Bus(logger)
        moves = [
            # heading zero
            [[Translation(0, 0, 0), Rotation(0, 0, 0, 1)], [[0, 0, 0], [0, 0, 0, 1]]],

            # looking left 90 degrees, moving front 1m
            # t265 turns around y axis +90, moving z -1
            # osgar turns round z axis +90, moving x 1
            [[Translation(0, 0, -1), Rotation(0, 0.7071068, 0, 0.7071068)], [[1, 0, 0], [0, 0, 0.7071068, 0.7071068]]],

            # looking up 90 degrees, moving up 1m
            # t265 turns around x axis +90, moving y 1
            # osgar turns round y axis -90, moving z 1
            [[Translation(0, 1, 0), Rotation(0.7071068, 0, 0, 0.7071068)], [[0, 0, 1], [0, -0.7071068, 0, 0.7071068]]]
        ]
        c = RealSense(bus=bus.handle('rs'), config={})
        tester = bus.handle('tester')
        bus.connect('rs.pose3d', 'tester.pose3d')
        for input, output in moves:
            frame = MagicMock()
            pose_frame = frame.as_pose_frame.return_value
            pose_frame.get_pose_data.return_value = Pose(
                Acceleration(0, 0, 0),
                AngularAcceleration(0, 0, 0),
                AngularVelocity(0, 0, 0),
                0,
                input[1],
                0,
                input[0],
                Velocity(0, 0, 0))
            frame.get_timestamp.return_value = 0
            frame.get_frame_number.return_value = 0
            c.pose_callback(frame)
            dt, channel, pose3d = tester.listen()
            self.assertEqual(channel, 'pose3d')
            self.assertEqual(pose3d, output)

    def test_depth(self):
        logger = MagicMock()
        logger.write = MagicMock(return_value=datetime.timedelta(microseconds=9721))
        bus = Bus(logger)
        c = RealSense(bus=bus.handle('rs'), config={"depth_rgb": True})
        tester = bus.handle('tester')
        bus.connect('rs.depth', 'tester.depth')
        bus.connect('rs.color', 'tester.color')
        frameset = MagicMock()
        frameset.is_frameset.return_value = True
        frame = frameset.as_frameset.return_value.get_depth_frame.return_value
        frame.get_timestamp.return_value = 0
        frame.get_frame_number.return_value = 0
        frame.is_depth_frame.return_value = True
        frame.as_depth_frame.return_value.get_data.return_value = [1,2]

        color_frame = frameset.as_frameset.return_value.get_color_frame.return_value
        color_frame.get_timestamp.return_value = 0
        color_frame.get_frame_number.return_value = 0
        color_frame.is_video_frame.return_value = True
        color_frame.as_video_frame.return_value.get_data.return_value = np.asarray([[0, 100, 255]], dtype=np.uint8)

        c.depth_callback(frameset)
        dt, channel_1, depth = tester.listen()
        dt, channel_2, color = tester.listen()
        depth_expected = np.asanyarray([1,2])
        color_expected = np.asanyarray([[0, 100, 255]], dtype=np.uint8)
        self.assertEqual(channel_1, 'depth')
        self.assertEqual(channel_2, 'color')
        color = cv2.imdecode(np.frombuffer(color, dtype=np.uint8), 0)
        self.assertEqual(depth.shape, depth_expected.shape)
        self.assertEqual(depth.dtype, depth_expected.dtype)
        self.assertTrue(np.array_equal(depth, depth_expected))
        self.assertEqual(color.shape, color_expected.shape)

# vim: expandtab sw=4 ts=4



"""
  Radio for Virtual SubT

"""
from datetime import timedelta

from osgar.node import Node
from osgar.bus import BusShutdownException
from osgar.lib.serialize import serialize, deserialize

# Note:
#  Payload size is limited to 1500 bytes - CommsClient::SendTo()

def draw_positions(arr):
    """
    Draw positions of tripples:
        (time, ID, (x. y, heading))
    """
    import matplotlib.pyplot as plt
    t = [a[0] for a in arr]
    robot_ids = sorted(list(set([a[1] for a in arr])))
    print('Robot IDs', robot_ids)

    for robot_id in robot_ids:
        x = [a[2][0]/1000.0 for a in arr if a[1] == robot_id]
        y = [a[2][1]/1000.0 for a in arr if a[1] == robot_id]
        line = plt.plot(x, y, '-o', linewidth=2, label=f'Robot {robot_id}')

#    plt.xlabel('time (s)')
    plt.axes().set_aspect('equal', 'datalim')
    plt.legend()
    plt.show()


class Radio(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register('radio', 'artf_xyz', 'breadcrumb', 'robot_xyz')
        self.min_transmit_dt = 0  # i.e. every sim_time_sec -> every simulated second
        self.last_transmit = None
        self.sim_time_sec = None
        self.recent_packets = []
        self.verbose = config.get('verbose', False)
        self.debug_robot_poses = []
        self.message_counter = 0

    def send_data(self, channel, data):
        raw = serialize([self.message_counter, channel, data])
        self.publish('radio', raw)
        self.message_counter += 1

    def on_radio(self, data):
        src, packet = data
        name = src.decode('ascii')
        __, channel, msg_data = deserialize(packet)
        if channel == 'artf':
            self.publish('artf_xyz', msg_data)  # topic rename - beware of limits 1500bytes!
        elif channel == 'breadcrumb':
            self.publish(channel, msg_data)
        elif channel == 'xyz':
            self.publish('robot_xyz', [name, msg_data])
            if self.verbose:
                self.debug_robot_poses.append((self.time, name, msg_data))

    def on_breadcrumb(self, data):
        self.send_data('breadcrumb', data)

    def on_pose3d(self, data):
        if self.sim_time_sec is None:
            return
        if self.last_transmit is None or self.sim_time_sec > self.last_transmit + self.min_transmit_dt:
            self.send_data('xyz', [self.sim_time_sec, data[0]])
            self.last_transmit = self.sim_time_sec

    def on_artf(self, data):
        self.send_data('artf', data)

    def on_sim_time_sec(self, data):
        self.sim_time_sec = data  # duplicate for super().update(), used just for easier unittesting
        self.send_data('sim_time_sec', data)

    def update(self):
        channel = super().update()  # define self.time
        handler = getattr(self, "on_" + channel, None)
        if handler is not None:
            handler(getattr(self, channel))
        else:
            assert False, channel  # not supported
        return channel

    def draw(self):
        """
        Debug Draw
        """
        draw_positions(self.debug_robot_poses)


# vim: expandtab sw=4 ts=4



import unittest
import logging
import math

from unittest.mock import MagicMock, call

from osgar.bus import Bus
from osgar.lib import quaternion
from subt.main import SubTChallenge
from subt.trace import distance3D

from subt import simulation

g_logger = logging.getLogger(__name__)


def entrance_reached(sim):
    corrected = [(rr - oo) for rr, oo in zip(sim.xyz, sim.origin)]
    goal = [0.5, 0, 0]  # note, that in real run the Y coordinate depends on choise left/righ
    if distance3D(corrected, goal) < 2:
        return True
    return False


class SubTChallengeTest(unittest.TestCase):

    def test_maybe_remember_artifact(self):
        config = {'max_speed': 0.5, 'walldist': 0.9, 'timeout': 600, 'symmetric': True,
                  'right_wall': True}
        bus = MagicMock()
        game = SubTChallenge(config, bus)

        artf_data = ['TYPE_BACKPACK', -1614, 1886]
        artf_xyz = (0, 0, 0)
        self.assertTrue(game.maybe_remember_artifact(artf_data, artf_xyz))

        # 2nd report should be ignored
        self.assertEqual(game.maybe_remember_artifact(artf_data, artf_xyz), False)

    def test_go_to_entrance(self):
        config = {'virtual_world': True, 'max_speed': 1.0, 'walldist': 0.8, 'timeout': 600, 'symmetric': False, 'right_wall': 'auto'}
        bus = Bus(simulation.SimLogger())
        app = SubTChallenge(config, bus.handle('app'))
        sim = simulation.Simulation(bus.handle('sim'), end_condition=entrance_reached)
        g_logger.info("connecting:")
        for o in bus.handle('sim').out:
            if o == 'pose2d':
                continue  # connect 'pose3d' only
            g_logger.info(f'  sim.{o} → app.{o}')
            bus.connect(f'sim.{o}', f'app.{o}')
        for o in bus.handle('app').out:
            g_logger.info(f'  app.{o} → sim.{o}')
            bus.connect(f'app.{o}', f'sim.{o}')
        g_logger.info("done.")
        app.start()
        sim.start()
        sim.join()
        app.request_stop()
        app.join()
        self.assertTrue(entrance_reached(sim))

    def test_on_artf_xyz(self):
        config = {'virtual_world': True, 'max_speed': 1.0, 'walldist': 0.8, 'timeout': 600, 'symmetric': False, 'right_wall': 'auto'}
        bus = MagicMock()
        app = SubTChallenge(config, bus)
        data = ['TYPE_BACKPACK', [5000, -2359, 2222]]
        app.xyz = 100.0, 2.0, 3.0
        app.orientation = quaternion.identity()
        bus.publish.reset_mock()
        app.on_artf(None, data)
        bus.publish.assert_called()
        self.assertEqual(bus.method_calls[-1],
                         call.publish('artf_xyz', [['TYPE_BACKPACK', [105000, -359, 5222], None, None]]))
        app.orientation = quaternion.euler_to_quaternion(math.pi/2, 0, 0)
        bus.publish.reset_mock()
        app.on_artf(None, data)
        bus.publish.assert_called()
        self.assertEqual(bus.method_calls[-1],
                         call.publish('artf_xyz', [['TYPE_BACKPACK', [102359, 7000, 5222], None, None]]))


if __name__ == "__main__":
    import sys
    logging.basicConfig(stream=sys.stderr, level=logging.DEBUG)
    unittest.main()

# vim: expandtab sw=4 ts=4




"""
  Replay driver modules from log file
"""

import argparse
import logging
from ast import literal_eval

from osgar import logger
from osgar.logger import LogReader
from osgar.lib.config import config_load, get_class_by_name
from osgar.bus import LogBusHandler, LogBusHandlerInputsOnly

g_logger = logging.getLogger(__name__)


def replay(args, application=None):
    log = LogReader(args.logfile, only_stream_id=0)
    print("original args:", next(log)[-1])  # old arguments
    config_str = next(log)[-1]
    config = literal_eval(config_str.decode('ascii'))
    if args.config is not None:
        config = config_load(*args.config, application=application)

    names = logger.lookup_stream_names(args.logfile)
    if args.debug:
        print("streams:")
        for i, name in enumerate(names):
            print(f" {i+1:2d} {name}")
    
    module = args.module
    assert module in config['robot']['modules'], (module, list(config['robot']['modules'].keys()))
    module_config = config['robot']['modules'][module]

    inputs = {}
    for edge_from, edge_to in config['robot']['links']:
        if edge_to.split('.')[0] == module:
            if edge_from not in names:
                g_logger.warning('Missing name: %s' % edge_from)
                names.append(edge_from)
            inputs[1 + names.index(edge_from)] = edge_to.split('.')[1]

    outputs = {i + 1: out.split('.')[1] for i, out in enumerate(names) if out.startswith(f"{module}.")}
    if args.debug:
        print("inputs:")
        for i, name in sorted(inputs.items()):
            print(f" {i:2d} {name}")
        print("outputs:")
        for i, name in sorted(outputs.items()):
            print(f" {i:2d} {name}")

    if args.force:
        reader = LogReader(args.logfile, only_stream_id=inputs.keys())
        bus = LogBusHandlerInputsOnly(reader, inputs=inputs)
    else:
        streams = list(inputs.keys()) + list(outputs.keys())
        reader = LogReader(args.logfile, only_stream_id=streams)
        bus = LogBusHandler(reader, inputs, outputs)

    driver_name = module_config['driver']
    module_class = get_class_by_name(driver_name)
    module_instance = module_class(module_config.get('init', {}), bus=bus)

    bus.node = module_instance # needed for slots
    return module_instance


def main():
    import signal
    logging.basicConfig(
        level=logging.DEBUG,
        format='%(name)-12s %(levelname)-8s %(message)s',
        #datefmt='%Y-%m-%d %H:%M',
    )
    logging.getLogger("matplotlib").setLevel(logging.INFO)
    parser = argparse.ArgumentParser(description='Replay module from log')
    parser.add_argument('logfile', help='recorded log file')
    parser.add_argument('--force', '-F', dest='force', action='store_true', help='force replay even for failing output asserts')
    parser.add_argument('--config', nargs='+', help='force alternative configuration file')
    parser.add_argument('--module', help='module name for analysis')  # TODO default "all"
    parser.add_argument('--verbose', '-v', help="verbose mode", action='store_true')
    parser.add_argument('--draw', help="draw debug results", action='store_true')
    parser.add_argument('--debug', help="print debug info about I/O streams", action='store_true')
    args = parser.parse_args()

    if args.module is None:
        modules = set()
        for s in logger.lookup_stream_names(args.logfile):
            modules.add(s.split('.')[0])
        print("Modules available for replay with `--module`:")
        for m in modules:
            print(" ", m)
        return

    module_instance = replay(args)
    module_instance.verbose = args.verbose

    signal.signal(signal.SIGINT, lambda signum, frame: module_instance.request_stop())
    module_instance.start()
    # now wait until the module is alive
    while module_instance.is_alive():
        module_instance.join(0.2)

    if not args.force:
        print("maximum delay:", module_instance.bus.max_delay)

    if args.draw:
        module_instance.draw()


if __name__ == "__main__":
    main()

# vim: expandtab sw=4 ts=4



import unittest
from unittest.mock import MagicMock
import math

from osgar.ro2018 import (geo_length, geo_angle,
                          latlon2xy, EmergencyStopMonitor, EmergencyStopException)


class RO2018Test(unittest.TestCase):

    def test_geo_lenght(self):
        self.assertAlmostEqual(
                geo_length((51749517, 180462688), (51749517, 180462688)),
                0.0)
        self.assertLess(
                geo_length((51749517, 180462688), (51749518, 180462688)),
                0.03)  # i.e. resolution ~ 3cm

    def test_geo_angle(self):
        self.assertAlmostEqual(
                geo_angle((51749517, 180462688), (51749617, 180462688)),
                0.0)  # go East
        self.assertAlmostEqual(
                geo_angle((51749517, 180462688), (51749517, 180462788)),
                math.radians(90.0))  # go North
        self.assertIsNone(
                geo_angle((51749517, 180462688), (51749518, 180462688)))
                # for too small distance

    def test_latlon2xy(self):
        self.assertEqual(latlon2xy(50.128246666666, 14.3748658333333), (51749517, 180462688 - 1000))

    def test_emergency_stop_monitor(self):
        robot = MagicMock()
        robot.register = MagicMock(return_value=42)
        with EmergencyStopMonitor(robot) as esm:
            robot.register.assert_called_once_with(esm.update)
            robot.status = 0x0
            with self.assertRaises(EmergencyStopException):
                esm.update(robot)
        robot.unregister.assert_called_once_with(42)

# vim: expandtab sw=4 ts=4



"""Artefacts detection via tensorflow"""

import os
import numpy as np
import cv2

# Directory with files saved_model.pb and saved_model.pbtxt
MODEL_DIR = "subt/tf_models"
# Files for CvDetector
PATH_TO_PB_GRAPH = "subt/tf_models/frozen_inference_graph.pb"
PATH_TO_CV_GRAPH = "subt/tf_models/cv_graph.pbtxt"

ARTF_NAME = np.array(["backpack", "survivor", "phone", "rope", "helmet", "robot", "breadcrumb", "nothing"])
MIN_SCORES = np.array([0.36, 0.25, 0.1, 0.1, 0.4, 1.0, 1.0, 1.0])


class CvDetector:
    def __init__(self):
        self.min_score = MIN_SCORES
        if os.path.exists(PATH_TO_PB_GRAPH) and os.path.exists(PATH_TO_CV_GRAPH):
            graph_path = PATH_TO_PB_GRAPH
            cv_graph = PATH_TO_CV_GRAPH
        else:  # docker
            graph_path = os.path.join(os.path.dirname(__file__), '../../../frozen_inference_graph.pb')
            cv_graph = os.path.join(os.path.dirname(__file__), '../../../cv_graph.pbtxt')

        self.cvNet = cv2.dnn.readNetFromTensorflow(graph_path, cv_graph)

    def subt_detector(self, img):
        ret  = []
        detection = self.detect(img)
        if detection is not None:
            num_detections = detection['num_detections']
            for ii in range(num_detections):
                name = detection['classes_names'][ii].item()
                score = detection['scores'][ii].item()
                bbox = detection['bboxes'][ii].tolist()
                ret.append([name, score, bbox])
        return ret

    def detect(self, img):
        rows = img.shape[0]
        cols = img.shape[1]
        # The size parameter should correspond with model
        # ssd_inception_v2
        self.cvNet.setInput(cv2.dnn.blobFromImage(img, size=(300, 300), swapRB=True, crop=False))
        # faster_rcnn_resnet50
#        self.cvNet.setInput(cv2.dnn.blobFromImage(img, size=(640, 320), swapRB=True, crop=False))
        cvOut = self.cvNet.forward()

        scores = cvOut[0, 0, :, 2]
        # ssd_inception_v2
        classes = cvOut[0, 0, :, 1]-1
        # faster_rcnn_resnet50
#        classes = cvOut[0, 0, :, 1]
        classes = classes.astype('int32')
        mask = self.min_score[classes] < scores

        num_detections = np.count_nonzero(mask)
        if num_detections > 0:
            scores = scores[mask]
            classes = classes[mask]
            bboxes = cvOut[0, 0, :, 3:][mask]
            bboxes = bboxes*np.array([cols, rows, cols, rows])
            bboxes = bboxes.astype('int32')
            classes_names = ARTF_NAME[classes]

            return { 'bboxes': bboxes, 'classes_names': classes_names, 'scores': scores, 'num_detections': num_detections }

        return None


    def run_on_image(self, img):
        detection = self.detect(img)
        if detection:
            num_detections = detection['num_detections']
            for ii in range(num_detections):
                x, y, w, h = detection['bboxes'][ii]
                score = detection['scores'][ii]
                name = detection['classes_names'][ii]
                cv2.rectangle(img, (x, y), (w, h), (255, 0, 0), thickness=2)
                cv2.putText(img, "%s %0.3f " % (name, score), (min(x, 520), max(20, y)), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255, 0, 0), 1,
                            cv2.LINE_AA)
        return img


class TfDetector:
    def __init__(self):
        import tensorflow as tf
        global tf
        self.model = tf.saved_model.load(MODEL_DIR)
        self.model = self.model.signatures['serving_default']
#        print(self.model.inputs)
        self.model.output_dtypes


    def detect(self, img):
        rows = img.shape[0]
        cols = img.shape[1]
        input_tensor = tf.convert_to_tensor(img)
        input_tensor = input_tensor[tf.newaxis, ...]
        raw_dict = self.model(input_tensor)

        scores = raw_dict['detection_scores'][0, :].numpy()
        classes = raw_dict['detection_classes'][0, :].numpy() - 1
        classes = classes.astype('int32')
        mask = MIN_SCORES[classes] < scores

        num_detections = np.count_nonzero(mask)
        if num_detections > 0:
            scores = scores[mask]
            classes = classes[mask]
            bboxes = raw_dict['detection_boxes'][0, :].numpy()[mask]
            bboxes = bboxes[:, [1, 0, 3, 2]]
            bboxes = bboxes*np.array([cols, rows, cols, rows])
            bboxes = bboxes.astype('int32')
            classes_names = ARTF_NAME[classes]

            return { 'bboxes': bboxes, 'classes_names': classes_names, 'scores': scores, 'num_detections': num_detections }

        return None


    def run_on_image(self, img):
        detection = self.detect(img)
        if detection:
            num_detections = detection['num_detections']
            for ii in range(num_detections):
                x, y, w, h = detection['bboxes'][ii]
                score = detection['scores'][ii]
                name = detection['classes_names'][ii]
                cv2.rectangle(img, (x, y), (w, h), (255, 0, 0), thickness=2)
                cv2.putText(img, "%s %0.3f " %(name, score), (min(x, 520), max(20, y)), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255, 0, 0), 1,
                            cv2.LINE_AA)
        return img


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('--path', help='path to log file, image or directory wit images')
    parser.add_argument('--stream', help='image stream', default='camera.raw')
    parser.add_argument('--cam', help='test with notebook camera', action='store_true')
    parser.add_argument('--tf', help='run detector with tensorflow', action='store_true')
    args = parser.parse_args()

    if args.tf:
        detector = TfDetector()
    else:
        detector = CvDetector()

    if args.cam:
        cam = cv2.VideoCapture(0)
        while True:
            __, img = cam.read()
            img = detector.run_on_image(img)
            cv2.imshow('img', img)
            k = cv2.waitKey(1) & 0xFF
            if k == ord("q"):
                cv2.destroyAllWindows()
                break
        cam.release()

    if args.path:
        path = args.path
        if os.path.isdir(path):
            im_file_list = os.listdir(path)
            for im_file in im_file_list:
                im_path = os.path.join(path, im_file)
                img = cv2.imread(im_path)
                if img is not None:
                    img = detector.run_on_image(img)
                    cv2.imshow('img', img)
                    k = cv2.waitKey(0) & 0xFF
                    if k == ord("q"):
                        cv2.destroyAllWindows()
                        break

        elif path.endswith(".log"):
            from osgar.logger import LogReader, lookup_stream_id
            from osgar.lib.serialize import deserialize
            stream_name = args.stream
            only_stream = lookup_stream_id(path, stream_name)
            with LogReader(path, only_stream_id=only_stream) as log:
                for timestamp, stream_id, data_raw in log:
                    buf_color = deserialize(data_raw)
                    img = cv2.imdecode(np.frombuffer(buf_color, dtype=np.uint8), 1)
                    img = detector.run_on_image(img)
                    cv2.imshow('img', img)
                    k = cv2.waitKey(200) & 0xFF
                    if k == ord("q"):
                        cv2.destroyAllWindows()
                        break

        else:
            img = cv2.imread(path)
            if img is not None:
                img = detector.run_on_image(img)
                cv2.imshow('img', img)
                cv2.waitKey(0)
                cv2.destroyAllWindows()
            else:
                print("no image")



import unittest
from unittest.mock import MagicMock
from datetime import timedelta

from osgar.drivers.cortexpilot import Cortexpilot
from osgar.bus import Bus

SAMPLE_DATA = bytes.fromhex(
        '000259010d0000000040952c416666be410000000000000000000000800000000000000' +
        '00000006fe702008013464840dd6e1db143f60900805f43ed44413d63e3ac3c3386edbb' +
        'aeb17f3f5d663943008037440000c040002079461ab754bd9c137d3cff149dbbf37345b' +
        'ed29ec8bbb763eabebf8304000000000000000000000000000000000000000000000000' +
        '00000000000000000000000000000000000000000000000000000000000000000000000' +
        '00000000000000000000000000000000000000000000000000000000000000000000000' +
        '00000000000000000000000000000000000000000000000000000000000000000000000' +
        '00000000000000000000000000000000000000000000000000000000000000000000000' +
        '00000000000000000000000000000000000000000000000000000000000000000000000' +
        '00000000000000000000000000000000000000000000000000000000000000000000000' +
        '00000000000000000000000000000000000000000000000000000000000000000000000' +
        '00000000000000000000000000000000000000000000000000000000000000000000000' +
        '00000000000000000000000000000000000000000000000000000000000000000000000' +
        '00000000000000000000000000000000000000000000000000000000000000000000000' +
        '00000000000000000000000000000000000000000000000000000000000000000000000' +
        '00000000000000000000000000000000000000000000000000000000000000000000000' +
        '000000000000000000000000000000000000000000000000000000000000000000000047')


class CortextpilotTest(unittest.TestCase):

    def test_usage(self):
        logger = MagicMock()
        logger.write = MagicMock(return_value=timedelta(seconds=135))
        bus = Bus(logger)
        handle = bus.handle('cortexpilot')
        tester = bus.handle('tester')
        robot = Cortexpilot(config={}, bus=handle)
        bus.connect('cortexpilot.raw', 'tester.raw')
        robot.start()
        robot.request_stop()
        robot.join()
        tester.shutdown()
        self.assertEqual(tester.listen(), (timedelta(seconds=135), 'raw', b'\x00\x00\x03\x01\x01\xfb'))

    def test_2nd_loop(self):
        logger = MagicMock()
        logger.write = MagicMock(return_value=timedelta(seconds=135))
        bus = Bus(logger)
        handle = bus.handle('cortexpilot')
        tester = bus.handle('tester')
        robot = Cortexpilot(config={}, bus=handle)
        bus.connect('cortexpilot.raw', 'tester.raw')
        handle.queue.put((timedelta(seconds=123), 'raw', b'\x00\x00\x10\x01\x01Robik V4.0.2\x00\x8f'))
        robot.start()
        robot.request_stop()
        robot.join()
        tester.shutdown() # so that listen() does not block

        self.assertEqual(tester.listen(), (timedelta(seconds=135), 'raw', b'\x00\x00\x03\x01\x01\xfb')) # request version
        self.assertEqual(tester.listen(), (timedelta(seconds=135), 'raw', bytes.fromhex('00000f010d000000000000000040010000a2'))) # cmd


    def test_create_packet(self):
        robot = Cortexpilot(config={}, bus=MagicMock())
        packet = robot.create_packet()
        self.assertEqual(len(packet), 3 + 15)
        self.assertEqual(sum(packet) % 256, 0)
        self.assertEqual(packet[-1], 0xa2)

        robot.desired_speed = 0.5
        robot.lidar_valid = True  # otherwise the speed will be reset to zero
        packet = robot.create_packet()
        self.assertEqual(len(packet), 3 + 15)
        self.assertEqual(sum(packet) % 256, 0)
        self.assertEqual(packet[-1], 99)

        # test packet checksum
        robot.desired_speed = -0.12314114151
        packet = robot.create_packet()

    def test_get_packet(self):
        robot = Cortexpilot(config={}, bus=MagicMock())
        packet = robot.get_packet()
        self.assertIsNone(packet)

        robot._buf = SAMPLE_DATA
        packet = robot.get_packet()
        self.assertIsNotNone(packet)
        self.assertEqual(len(packet), len(SAMPLE_DATA))
        self.assertEqual(len(robot._buf), 0)

        packet = robot.get_packet()
        self.assertIsNone(packet)

    def test_parse_packet(self):
        robot = Cortexpilot(config={}, bus=MagicMock())
        robot.parse_packet(SAMPLE_DATA)
        self.assertEqual(robot.flags, 0x0)
        self.assertAlmostEqual(robot.voltage[0], 10.78643798)
        self.assertAlmostEqual(robot.voltage[1], 23.79999923)

        # trigger pose update
        robot.parse_packet(SAMPLE_DATA)

# vim: expandtab sw=4 ts=4



import unittest
import math

from osgar.drivers.simulator import SpiderSimulator


class SpiderSimulatorTest(unittest.TestCase):

    def test_usage(self):
        config = {'position': [51748232, 180462051]}
        bus = None  # BusHandler(logger)
        spider = SpiderSimulator(config, bus)

        spider.desired_speed = 0
        spider.desired_wheel_angle = 0
        spider.step(0.1)
        self.assertEqual(spider.rel_pose, (0, 0, 0))

        spider.desired_speed = 10
        spider.desired_wheel_angle = 0
        spider.step(2.0)
        self.assertGreater(spider.rel_pose[0], 0.0)

        spider.desired_speed = 1
        spider.desired_wheel_angle = math.radians(45)
        spider.status = 0x3
        spider.step(1.0)
        self.assertGreater(spider.wheel_angle, 0.0)
        self.assertEqual(spider.status, 0x8003)  # alive bit
        spider.step(1.0)
        self.assertEqual(spider.status, 0x3)

    def test_get_position(self):
        config = {'position': [51748232, 180462051]}
        spider = SpiderSimulator(config=config, bus=None)
        p = spider.get_position()
        self.assertEqual(p, [51748232, 180462051])

        spider.rel_pose = (100, 0, 0)  # 100m to East
        p = spider.get_position()
        self.assertEqual(p, [51748232 + 3344, 180462051])

        spider.rel_pose = (0, -1, 0)  # 1m to South
        p = spider.get_position()
        self.assertEqual(p, [51748232, 180462051 - 33])

        spider.rel_pose = (100, 0, 0)  # 100m to East on eqator
        spider.ref_position = [0, 180462051]
        p = spider.get_position()
        self.assertEqual(p, [3240, 180462051])

    def test_get_wheels_angles_raw(self):
        config = {'position': [51748232, 180462051]}
        spider = SpiderSimulator(config=config, bus=None)
        angles = spider.get_wheels_angles_raw()
        self.assertEqual(angles, [0, 0, 0, 0])

        spider.wheel_angle = math.radians(270)
        angles = spider.get_wheels_angles_raw()
        self.assertEqual(angles, [128, 128, 128, 128])

# vim: expandtab sw=4 ts=4



import unittest
import os
import tempfile
import time
import logging

from threading import Timer
from datetime import datetime, timedelta, timezone
from unittest.mock import patch
from contextlib import ExitStack
from pathlib import Path

import osgar.logger  # needed for patching the osgar.logger.datetime.datetime
from osgar.logger import (LogWriter, LogReader, LogAsserter, INFO_STREAM_ID,
                          lookup_stream_id, LogIndexedReader)



class TimeStandsStill:
    # inspired by
    # https://marcusahnve.org/blog/2017/mocking-datetime-in-python/
    def __init__(self, datetime):
        self.datetime = datetime.replace(tzinfo=timezone.utc)

    def now(self, tzinfo):
        assert tzinfo == timezone.utc
        return self.datetime


def delayed_copy(src, dst, skip_size):
    with open(src, 'rb') as f:
        buf = f.read(skip_size)
        assert len(buf) == skip_size, (len(buf), skip_size)
        with open(dst, 'ab+') as f2:
            buf = f.read()
            f2.write(buf)
            f2.flush()


class LoggerStreamingTest(unittest.TestCase):

    def setUp(self):
        ref = os.environ.get('OSGAR_LOGS')
        if ref is None:
            os.environ['OSGAR_LOGS'] = "."

    def test_timedelta_sequence(self):
        parse_timedelta = osgar.logger.timedelta_parser()
        for i in range(20000):
            t_ref = timedelta(seconds=i)
            binary = osgar.logger.format_timedelta(t_ref)
            t_parsed = parse_timedelta(binary)
            self.assertEqual(t_ref, t_parsed)

    def test_timedelta_cases(self):
        t_ref = timedelta(hours=1, minutes=11, seconds=35)
        last_dt = t_ref - timedelta(seconds=1)
        binary = osgar.logger.format_timedelta(t_ref)
        parse_timedelta = osgar.logger.timedelta_parser(last_dt)
        t_parsed = parse_timedelta(binary)
        self.assertEqual(t_ref, t_parsed)

    def test_writer_prefix(self):
        with LogWriter(prefix='tmp') as log:
            self.assertTrue(Path(log.filename).name.startswith('tmp'))
        os.remove(log.filename)

    def test_context_manager(self):
        with LogWriter(prefix='tmpp', note='1st test') as log:
            filename = log.filename
            start_time = log.start_time
            t1 = log.write(10, b'\x01\x02\x02\x04')
            time.sleep(0.01)
            t2 = log.write(10, b'\x05\x06\x07\x08')
            self.assertLess(t1, t2)
        
        with LogReader(filename) as log:
            self.assertEqual(start_time, log.start_time)

            __, stream_id, data = next(log)
            self.assertEqual(INFO_STREAM_ID, stream_id)

            t, stream_id, data = next(log)
            self.assertEqual(stream_id, 10)
            self.assertEqual(data, b'\x01\x02\x02\x04')

            t, stream_id, data = next(log)
            self.assertTrue(t.microseconds > 100)
            
            with self.assertRaises(StopIteration):
                __ = next(log)

        with LogReader(filename, only_stream_id=10) as log:
            for t, stream_id, data in log:
                self.assertEqual(stream_id, 10)
                self.assertEqual(data, b'\x01\x02\x02\x04')
                break

        os.remove(log.filename)

    def test_read_two_streams(self):
        with LogWriter(prefix='tmp2', note='test_read_two_streams') as log:
            filename = log.filename
            t1 = log.write(1, b'\x01\x02\x02\x04')
            time.sleep(0.001)
            t2 = log.write(3, b'\x05\x06')
            time.sleep(0.001)
            t3 = log.write(2, b'\x07\x08')

        with LogReader(filename, only_stream_id=[1, 2]) as log:
            arr = []
            for t, stream_id, data in log:
                self.assertIn(stream_id, [1, 2])
                arr.append((t, stream_id))
            self.assertEqual(arr, [(t1, 1), (t3, 2)])

        os.remove(log.filename)

    def test_register(self):
        with LogWriter(prefix='tmp2', note='test_register') as log:
            filename = log.filename
            self.assertEqual(log.register('raw'), 1)

            with self.assertRaises(AssertionError):
                log.register('raw')  # duplicity name

            self.assertEqual(log.register('gps.position'), 2)

        with LogReader(filename, only_stream_id=INFO_STREAM_ID) as log:
            arr = []
            for __, __, data in log:
                if b'names' in data:
                    arr.append(data)
            self.assertEqual(len(arr), 2, arr)
            self.assertEqual(arr[0], b"{'names': ['raw']}")
            self.assertEqual(arr[1], b"{'names': ['raw', 'gps.position']}")

        os.remove(filename)

    def test_log_asserter(self):
        with LogWriter(prefix='tmp3', note='test_log_asserter') as log:
            filename = log.filename
            t1 = log.write(1, b'\x01\x02')
            time.sleep(0.001)
            t2 = log.write(2, b'\x05')
            time.sleep(0.001)
            t3 = log.write(1, b'\x07\x08')

        with LogAsserter(filename, only_stream_id=[1, 2]) as log:
            log.assert_stream_id = 2
            arr = []
            for t, stream_id, data in log:
                self.assertIn(stream_id, [1,2])
                if stream_id == 2:
                    log.write(2, b'\x05')
                arr.append((t, stream_id))
            self.assertEqual(arr, [(t1, 1), (t2, 2), (t3, 1)])

        os.remove(log.filename)

    def test_lookup_stream_id(self):
        self.assertEqual(lookup_stream_id('dummy.log', None), None)
        self.assertEqual(lookup_stream_id('dummy.log', '3'), 3)
        # TODO name lookup

    def test_large_block(self):
        data = bytes([x for x in range(100)]*1000)
        self.assertEqual(len(data), 100000)
        with LogWriter(prefix='tmp4', note='test_large_block') as log:
            filename = log.filename
            t1 = log.write(1, data)
            t2 = log.write(1, data[:0xFFFF])
            t3 = log.write(1, b'')
            t4 = log.write(1, b'ABC')
            t5 = log.write(1, data+data)  # multiple split

        with LogReader(filename, only_stream_id=1) as log:
            arr = []
            for __, __, data in log:
                arr.append(data)
            self.assertEqual([len(x) for x in arr],
                             [100000, 65535, 0, 3, 200000])

        os.remove(log.filename)

    def test_environ(self):
        with LogWriter(prefix='tmp5', note='test_filename_before') as log:
            self.assertTrue(Path(log.filename).name.startswith('tmp5'))
        os.remove(log.filename)

        os.environ['OSGAR_LOGS'] = 'tmp_dir'
        with LogWriter(prefix='tmp6', note='test_filename_after') as log:
            self.assertEqual(Path(log.filename).parent.name, 'tmp_dir')
        os.remove(log.filename)

        del os.environ['OSGAR_LOGS']
        with self.assertLogs(level=logging.WARNING):
            with LogWriter(prefix='tmp7', note='test_filename_after2') as log:
                self.assertTrue(Path(log.filename).name.startswith('tmp7'))
        os.remove(log.filename)

    def test_time_overflow(self):
        with LogWriter(prefix='tmp8', note='test_time_overflow') as log:
            log.start_time = datetime.now(timezone.utc) - timedelta(hours=1, minutes=30)
            t1 = log.write(1, b'\x01\x02')
            self.assertGreater(t1, timedelta(hours=1))
            filename = log.filename
        with LogReader(filename, only_stream_id=1) as log:
            dt, channel, data = next(log)
            self.assertGreater(dt, timedelta(minutes=10))
        os.remove(filename)


    def test_time_overflow2(self):
        with patch('osgar.logger.datetime.datetime'):
            osgar.logger.datetime.datetime = TimeStandsStill(datetime(2020, 1, 21))
            with osgar.logger.LogWriter(prefix='tmp9', note='test_time_overflow') as log:
                filename = log.filename
                t1 = log.write(1, b'\x01')
                self.assertEqual(t1, timedelta(0))
                osgar.logger.datetime.datetime = TimeStandsStill(datetime(2020, 1, 21, 1))
                t2 = log.write(1, b'\x02')
                self.assertEqual(t2, timedelta(hours=1))
                osgar.logger.datetime.datetime = TimeStandsStill(datetime(2020, 1, 21, 2))
                t3 = log.write(1, b'\x03')
                self.assertEqual(t3, timedelta(hours=2))
                osgar.logger.datetime.datetime = TimeStandsStill(datetime(2020, 1, 21, 4))
                # TODO this write should rise exception as the time gap is too big to track!
                t4 = log.write(1, b'\x04')
                self.assertEqual(t4, timedelta(hours=4))
        with LogReader(filename, only_stream_id=1) as log:
            dt, channel, data = next(log)
            self.assertEqual(dt, timedelta(hours=0))
            dt, channel, data = next(log)
            self.assertEqual(dt, timedelta(hours=1))
            dt, channel, data = next(log)
            self.assertEqual(dt, timedelta(hours=2))
#            dt, channel, data = next(log)
#            self.assertEqual(dt, timedelta(hours=4))
        os.remove(filename)

    def test_large_blocks_with_time_overflow(self):
        with patch('osgar.logger.datetime.datetime'):
            osgar.logger.datetime.datetime = TimeStandsStill(datetime(2019, 1, 1))
            with osgar.logger.LogWriter(prefix='tmpA', note='test_time_overflow with large blocks') as log:
                filename = log.filename
                t1 = log.write(1, b'\x01'*100000)
                self.assertEqual(t1, timedelta(0))
                osgar.logger.datetime.datetime = TimeStandsStill(datetime(2019, 1, 1, 1))
                t2 = log.write(1, b'\x02'*100000)
                self.assertEqual(t2, timedelta(hours=1))
                osgar.logger.datetime.datetime = TimeStandsStill(datetime(2019, 1, 1, 2))
                t3 = log.write(1, b'\x03'*100000)
                self.assertEqual(t3, timedelta(hours=2))
        with LogReader(filename, only_stream_id=1) as log:
            dt, channel, data = next(log)
            self.assertEqual(dt, timedelta(hours=0))
            dt, channel, data = next(log)
            self.assertEqual(dt, timedelta(hours=1))
            dt, channel, data = next(log)
            self.assertEqual(dt, timedelta(hours=2))
        os.remove(filename)

    def test_no_eof(self):
        # similar to tail -f (--follow)
        with LogWriter(prefix='tmpB', note='test_EOF') as log:
            filename = log.filename
            t1 = log.write(1, b'\x01'*100)
            time.sleep(0.001)
            t2 = log.write(1, b'\x02'*100)
            time.sleep(0.001)
            t3 = log.write(1, b'\x03'*100000)

        with LogReader(filename, only_stream_id=1) as log:
            dt, channel, data = next(log)
            self.assertEqual(data, b'\x01'*100)
            dt, channel, data = next(log)
            self.assertEqual(data, b'\x02'*100)

        partial = filename + '.part'
        with open(filename, 'rb') as f_in:
            with open(partial, 'wb') as f_out:
                f_out.write(f_in.read(100))
        
        with LogReader(partial, only_stream_id=1) as log:
            with self.assertLogs(osgar.logger.__name__, logging.ERROR):
                with self.assertRaises(StopIteration):
                    dt, channel, data = next(log)

        proc = Timer(0.1, delayed_copy, [filename, partial, 100])
        proc.start()
        with LogReader(partial, follow=True, only_stream_id=1) as log:
            dt, channel, data = next(log)
            self.assertEqual(data, b'\x01'*100)
        proc.join()

        partial = filename + '.part'
        with open(filename, 'rb') as f_in:
            with open(partial, 'wb') as f_out:
                f_out.write(f_in.read(100000))

        with LogReader(partial, only_stream_id=1) as log:
            dt, channel, data = next(log)
            self.assertEqual(data, b'\x01'*100)
            dt, channel, data = next(log)
            self.assertEqual(data, b'\x02'*100)
            with self.assertRaises(AssertionError):
                dt, channel, data = next(log)

        os.remove(partial)
        os.remove(filename)

    def test_read_gen_seek(self):
        with LogWriter(prefix='tmpC', note='test_read_gen_seek') as log:
            filename = log.filename
            t1 = log.write(1, b'\x01')
            time.sleep(0.001)
            t2 = log.write(1, b'\x02')
            time.sleep(0.001)
            t3 = log.write(1, b'\x03')

        with LogReader(filename, only_stream_id=1) as log:
            dt, channel, data = next(log)
            self.assertEqual(data, b'\x01')
            dt, channel, data = next(log)
            self.assertEqual(data, b'\x02')
        os.remove(filename)

    def test_large_invalid(self):
        header = osgar.logger.format_header(datetime(2019, 1, 1, 1))
        packet = osgar.logger.format_packet(1, b"\x00"*(2**16), timedelta())
        invalid = b"\xFF"*len(packet[-2]) # overwrite header of the last subpacket
        self.assertNotEqual(packet[-2], invalid)
        packet[-2] = invalid
        logdata = b"".join(header + packet)

        with tempfile.TemporaryDirectory(dir=".") as d:
            filename = Path(".") / d / "log"
            with open(filename, "wb") as f:
                f.write(logdata[:-1])
            with LogReader(filename) as l:
                with self.assertRaises(StopIteration),\
                     self.assertLogs(logger=osgar.logger.__name__, level=logging.ERROR):
                    next(l)

    def test_incomplete(self):
        header = osgar.logger.format_header(datetime(2019, 1, 1, 1))
        packet = osgar.logger.format_packet(1, b"\x00"*10, timedelta())
        logdata = b"".join(header + packet)

        with tempfile.TemporaryDirectory(dir=".") as d:
            filename = Path(".") / d / "log"
            with open(filename, "wb") as f:
                f.write(logdata[:-1])
            with LogReader(filename) as l:
                with self.assertLogs(osgar.logger.__name__, logging.ERROR):
                    with self.assertRaises(StopIteration):
                        next(l)


class LoggerIndexedTest(unittest.TestCase):

    def setUp(self):
        ref = os.environ.get('OSGAR_LOGS')
        if ref is None:
            os.environ['OSGAR_LOGS'] = "."

    def test_indexed_reader(self):
        note = 'test_indexed_reader'
        sample = [b'\x01', b'\x02', b'\x03']
        times = []
        with ExitStack() as at_exit:

            with LogWriter(prefix='tmpIndexed', note=note) as log:
                at_exit.callback(os.remove, log.filename)
                for a in sample:
                    t = log.write(1, a)
                    times.append(t)
                    time.sleep(0.001)

            with LogIndexedReader(log.filename) as log:
                dt, channel, data = log[0]
                assert data.decode('utf-8') == note, data
                for i in range(len(sample)):
                    dt, channel, data = log[i+1]
                    assert data == sample[i], data
                    assert dt == times[i], (dt, times[i])
                assert len(log) == len(sample) + 1, len(log)

    def test_indexed_large_block(self):
        data = bytes([x for x in range(100)]*1000)
        self.assertEqual(len(data), 100000)
        with ExitStack() as at_exit:
            with LogWriter(prefix='tmpIndexedLarge', note='test_large_block') as log:
                at_exit.callback(os.remove, log.filename)
                t1 = log.write(1, data)
                t2 = log.write(1, data[:0xFFFF])
                t3 = log.write(1, b'')
                t4 = log.write(1, b'ABC')
                t5 = log.write(1, data+data)  # multiple split

            with LogIndexedReader(log.filename) as log:
                self.assertEqual(len(log[1][2]), 100000)
                self.assertEqual(len(log[2][2]), 65535)
                self.assertEqual(len(log[3][2]), 0)
                self.assertEqual(len(log[4][2]), 3)
                self.assertEqual(len(log[5][2]), 200000)

    def test_time_overflow(self):
        with ExitStack() as at_exit:
            with patch('osgar.logger.datetime.datetime'):
                osgar.logger.datetime.datetime = TimeStandsStill(datetime(2020, 1, 21))
                with osgar.logger.LogWriter(prefix='tmp9', note='test_time_overflow') as log:
                    at_exit.callback(os.remove, log.filename)
                    t1 = log.write(1, b'\x01')
                    self.assertEqual(t1, timedelta(0))
                    osgar.logger.datetime.datetime = TimeStandsStill(datetime(2020, 1, 21, 1))
                    t2 = log.write(1, b'\x02')
                    self.assertEqual(t2, timedelta(hours=1))
                    osgar.logger.datetime.datetime = TimeStandsStill(datetime(2020, 1, 21, 2))
                    t3 = log.write(1, b'\x03')
                    self.assertEqual(t3, timedelta(hours=2))
                    osgar.logger.datetime.datetime = TimeStandsStill(datetime(2020, 1, 21, 4))
                    # TODO this write should rise exception as the time gap is too big to track!
                    t4 = log.write(1, b'\x04')
                    self.assertEqual(t4, timedelta(hours=4))
            with LogIndexedReader(log.filename) as log:
                dt, channel, data = log[1]
                self.assertEqual(dt, timedelta(hours=0))
                dt, channel, data = log[2]
                self.assertEqual(dt, timedelta(hours=1))
                dt, channel, data = log[3]
                self.assertEqual(dt, timedelta(hours=2))
    #            dt, channel, data = next(log)
    #            self.assertEqual(dt, timedelta(hours=4))

    def test_large_blocks_with_time_overflow(self):
        with ExitStack() as at_exit:
            with patch('osgar.logger.datetime.datetime'):
                osgar.logger.datetime.datetime = TimeStandsStill(datetime(2019, 1, 1))
                with osgar.logger.LogWriter(prefix='tmpA', note='test_time_overflow with large blocks') as log:
                    at_exit.callback(os.remove, log.filename)
                    t1 = log.write(1, b'\x01'*100000)
                    self.assertEqual(t1, timedelta(0))
                    osgar.logger.datetime.datetime = TimeStandsStill(datetime(2019, 1, 1, 1))
                    t2 = log.write(1, b'\x02'*100000)
                    self.assertEqual(t2, timedelta(hours=1))
                    osgar.logger.datetime.datetime = TimeStandsStill(datetime(2019, 1, 1, 2))
                    t3 = log.write(1, b'\x03'*100000)
                    self.assertEqual(t3, timedelta(hours=2))
            with LogIndexedReader(log.filename) as log:
                dt, channel, data = log[1]
                self.assertEqual(dt, timedelta(hours=0))
                dt, channel, data = log[2]
                self.assertEqual(dt, timedelta(hours=1))
                dt, channel, data = log[3]
                self.assertEqual(dt, timedelta(hours=2))

    def test_no_eof(self):
        with patch('osgar.logger.datetime.datetime'):
            osgar.logger.datetime.datetime = TimeStandsStill(datetime(2019, 1, 1))
            with LogWriter(prefix='tmpEof', note='test_EOF') as log:
                filename = log.filename
                osgar.logger.datetime.datetime = TimeStandsStill(datetime(2019, 1, 1, 1))
                t1 = log.write(1, b'\x01'*100)
                osgar.logger.datetime.datetime = TimeStandsStill(datetime(2019, 1, 1, 2))
                t2 = log.write(1, b'\x02'*100)
                osgar.logger.datetime.datetime = TimeStandsStill(datetime(2019, 1, 1, 3))
                t3 = log.write(1, b'\x03'*100000)

        partial = filename + '.part'
        with open(filename, 'rb') as f_in, open(partial, 'wb') as f_out:
            f_out.write(f_in.read(100))
            f_out.flush()

            with LogIndexedReader(partial) as log:
                self.assertEqual(len(log), 1)
                dt, channel, data = log[0]
                with self.assertRaises(IndexError):
                    log[1]

                f_out.write(f_in.read(100))
                f_out.flush()
                log.grow()
                self.assertEqual(len(log), 2)
                dt, channel, data = log[1]
                self.assertEqual(data, b'\x01'*100)
                self.assertEqual(dt, timedelta(hours=1))
                with self.assertRaises(IndexError):
                    log[2]

                f_out.write(f_in.read())
                f_out.flush()

                log.grow()
                dt, channel, data = log[2]
                self.assertEqual(dt, timedelta(hours=2))
                self.assertEqual(data, b'\x02'*100)
                dt, channel, data = log[3]
                self.assertEqual(dt, timedelta(hours=3))
                self.assertEqual(len(log), 4)
                self.assertEqual(data, b'\x03'*100000)

        os.remove(partial)
        os.remove(filename)

    def test_large_blocks_with_growing_file(self):
        block_size = 100000 # fits into 2 packets, so 16 bytes overhead
        with ExitStack() as at_exit:
            with patch('osgar.logger.datetime.datetime'):
                osgar.logger.datetime.datetime = TimeStandsStill(datetime(2019, 1, 1))
                with osgar.logger.LogWriter(prefix='tmpA', note='') as log:
                    at_exit.callback(os.remove, log.filename)
                    t1 = log.write(1, b'\x01'*block_size)
                    self.assertEqual(t1, timedelta(0))
                    osgar.logger.datetime.datetime = TimeStandsStill(datetime(2019, 1, 1, 1))
                    t2 = log.write(1, b'\x02'*block_size)
                    self.assertEqual(t2, timedelta(hours=1))
                    osgar.logger.datetime.datetime = TimeStandsStill(datetime(2019, 1, 1, 2))
                    t3 = log.write(1, b'\x03'*block_size)
                    self.assertEqual(t3, timedelta(hours=2))

            partial = log.filename + '.part'
            with open(log.filename, 'rb') as f_in, open(partial, 'wb') as f_out:
                at_exit.callback(os.remove, partial)
                # log file starts with 16 byte header
                f_out.write(f_in.read(16))
                f_out.write(f_in.read(100))
                f_out.flush()

                with LogIndexedReader(partial) as log:
                    with self.assertRaises(IndexError):
                        log[0]
                    # add whole block plus 2 headers minus 100 already there
                    f_out.write(f_in.read(block_size-100+16))
                    f_out.flush()
                    log.grow()
                    self.assertEqual(len(log), 1)

                    # add a partial header only
                    f_out.write(f_in.read(4))
                    f_out.flush()
                    log.grow()
                    self.assertEqual(len(log), 1)

                    # add another block
                    f_out.write(f_in.read(block_size+16))
                    f_out.flush()
                    log.grow()

                    dt, channel, data = log[0]
                    self.assertEqual(dt, timedelta(hours=0))
                    dt, channel, data = log[1]
                    self.assertEqual(dt, timedelta(hours=1))

                    f_out.write(f_in.read())
                    f_out.flush()
                    log.grow()

                    dt, channel, data = log[2]
                    self.assertEqual(dt, timedelta(hours=2))

    def test_large_invalid(self):
        header = osgar.logger.format_header(datetime(2019, 1, 1, 1))
        packet = osgar.logger.format_packet(1, b"\x00"*(2**16), timedelta())
        invalid = b"\xFF"*len(packet[-2])
        self.assertNotEqual(packet[-2], invalid)
        packet[-2] = invalid
        logdata = b"".join(header + packet)

        with tempfile.TemporaryDirectory(dir=".") as d:
            filename = Path(".") / d / "log"
            with open(filename, "wb") as f:
                f.write(logdata[:-1])
            with self.assertLogs(logger=osgar.logger.__name__, level=logging.ERROR) as log:
                with LogIndexedReader(filename) as l:
                    with self.assertRaises(IndexError):
                        l[0]

    def test_incomplete(self):
        header = osgar.logger.format_header(datetime(2019, 1, 1, 1))
        packet = osgar.logger.format_packet(1, b"\x00"*10, timedelta())
        logdata = b"".join(header + packet)

        with tempfile.TemporaryDirectory(dir=".") as d:
            filename = Path(".") / d / "log"
            with open(filename, "wb") as f:
                f.write(logdata[:-1])
            with LogIndexedReader(filename) as l:
                with self.assertRaises(IndexError):
                    l[0]


# vim: expandtab sw=4 ts=4



"""
  Detect SubT Artifact in Camera Image
"""
from datetime import timedelta
import os
import tempfile
from io import StringIO

import cv2
import numpy as np

from osgar.node import Node
from osgar.bus import BusShutdownException


EXTINGUISHER = 'TYPE_EXTINGUISHER'
BACKPACK = 'TYPE_BACKPACK'
VALVE = 'TYPE_VALVE'
ELECTRICAL_BOX = 'TYPE_ELECTRICAL_BOX'
PHONE = 'TYPE_PHONE'
RADIO = 'TYPE_RADIO'
TOOLBOX = 'TYPE_TOOLBOX'
DUCT = 'TYPE_DUCT'
DRILL = 'TYPE_DRILL'
RESCUE_RANDY = 'TYPE_RESCUE_RANDY'
VENT = 'TYPE_VENT'
GAS = 'TYPE_GAS'
HELMET = 'TYPE_HELMET'
ROPE = 'TYPE_ROPE'

RED_THRESHOLD = 100 #300  # for close backpacks was good 1000  # virtual QVGA=50, used to be 100, urban=1000
YELLOW_THRESHOLD = 200  #500  # was 80
YELLOW_MAX_THRESHOLD = 4000  # 2633 known example
RED_YELLOW_MIN_3D_THRESHOLD = 50  # number of colored pixels in given depth distance threshold
WHITE_THRESHOLD = 20000

CO2_REPORT_LIMIT = 800
PHONE_PREFIX = "PhoneArtifact"
PHONE_SIGNAL_LIMIT = -50


g_mask = None

def old_count_red(img):
    count = 0
    for x in range(320):
        for y in range(240):
            b, g, r = img[y][x]
            if r > 100 and r > 2 * g and r > 2 * b:
                count += 1
    return count


def count_mask(mask):
    """Count statistics and bounding box for given image mask"""
    count = int(mask.sum())
    if count == 0:
        return count, None, None, None, None

    # argmax for mask finds the first True value
    x_min = (mask.argmax(axis=0) != 0).argmax()
    x_max = mask.shape[1] - np.flip((mask.argmax(axis=0) != 0), axis=0).argmax() - 1
    w = (mask.shape[1] - np.flip((mask.argmax(axis=0) != 0), axis=0).argmax()
            - (mask.argmax(axis=0) != 0).argmax())
    h = (mask.shape[0] - np.flip((mask.argmax(axis=1) != 0), axis=0).argmax()
            - (mask.argmax(axis=1) != 0).argmax())
    return count, w, h, x_min, x_max


def count_red(img, filtered=False, stdout=None):
    b = img[:,:,0]
    g = img[:,:,1]
    r = img[:,:,2]
    mask = np.logical_and(r > 50, np.logical_and(r/3 > g, r/3 > b))  # QVGA virtual, dark images, used to be 100, r/2
                                                                     # 3/4-VGA urban, light, strictly red r/3
    not_mask = np.logical_not(mask)
    kernel = np.ones((2,2), np.uint8)
    img2 = img.copy()
    img2[mask] = (255, 255, 255)
    img2[not_mask] = (0, 0, 0)
    img2 = cv2.erode(img2, kernel, iterations=1)
#    cv2.imwrite('artf.jpg', img2)
#        print('mask', mask.shape)
#        for x in range(mask.shape[0]):
#            for y in range(mask.shape[1]):
#                if mask[x][y]:
#                    print(x, y)
    b = img2[:,:,0]
    mask = b == 255
    global g_mask
    g_mask = mask.copy()
    if filtered:
        # detect largest blob in img2
        kernel = np.ones((3,3), np.uint8)
        img2 = cv2.dilate(img2, kernel, iterations=1)
#        cv2.imwrite('artf.jpg', img2)
        b = img2[:,:,0]
        mask = b == 255
        reddish = b.astype(np.uint8)
        __, bin_img = cv2.threshold(reddish, 160, 255, cv2.THRESH_BINARY)

        # cv2.findCountours for opencv == 3 returns 3 values while for opencv == 4 returns just 2
        contours, hierarchy = cv2.findContours(bin_img, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)[-2:]
        best_cnt = None
        best_area = None
        for cnt in contours:
            area = cv2.contourArea(cnt)
            if best_area is None or area > best_area:
                best_area = area
                best_cnt = cnt
#        print(best_area, best_cnt)
        x, y, w, h = cv2.boundingRect(best_cnt)

#        if stdout is not None:
#            # dump image to stdout
#            # https://stackoverflow.com/questions/50670326/how-to-check-if-point-is-placed-inside-contour
#            print(x, y, w, h)
#            for j in range(y, y + h):
#                s = ''
#                for i in range(x, x + w ):
#                    s += 'X' if mask[j][i] else '.'
#                stdout(s)
        # count, w, h, x_min, x_max
        return int(best_area), w, h, x, x+w

    return count_mask(mask)


def count_white(img):
    b = img[:,:,0]
    g = img[:,:,1]
    r = img[:,:,2]
    mask = np.logical_and(r >= 250, np.logical_and(g >= 250, b >= 250))
    return count_mask(mask)


def count_yellow(img):
    b = img[:,:,0]
    g = img[:,:,1]
    r = img[:,:,2]
    mask = np.logical_and(np.logical_and(r >= 60, g*0.95 > r), np.logical_and(r*0.5 > b, g*0.5 > b))
    global g_mask
    g_mask = mask.copy()
    # debug
#    img2 = img.copy()
#    img2[mask] = (0, 0, 255)
#    cv2.imwrite('artf.jpg', img2)
    return count_mask(mask)


def count_orange_blue(img):  # for phone
    b = img[:,:,0]
    g = img[:,:,1]
    r = img[:,:,2]
    mask_blue = np.logical_and(b > 40, np.logical_and(b/1.5 > g, b/2 > r))

    # detect phone only in the lowest 1/4 of the image (the phone is lying on the ground)
    blue = count_mask(mask_blue[90:,:])
    count, w, h, x_min, x_max = blue
    if count == 0 or w > 20 or h > 20:
        return 0, None, None, None, None
#    print(count)
    mask_orange = np.logical_and(r > 40, np.logical_and(r/1.5 > g, r/2 > b))
    orange = count_mask(mask_orange)

#    not_mask = np.logical_not(mask_orange)
#    img2 = img.copy()
#    img2[not_mask] = (0, 0, 0)
#    img2[mask_orange] = (0, 128, 255)
#    img2[mask_blue] = (255, 0, 0)
#    cv2.imwrite('artf.jpg', img2)

    mask = np.logical_or(mask_orange, mask_blue)
    return blue  #count_mask(mask)


def artf_in_scan(scan, width, img_x_min, img_x_max, verbose=False):
    """return precise artefact angle and distance for lidar & camera combination"""
    if scan is None:
        return 0, 0
    # the scan is already in mm, so angle is modified to int deg*100, ready to send
    x_min, x_max = img_x_min, img_x_max

    angular_resolution = len(scan) / 270
    mid_index = len(scan) // 2
    camera_fov_deg = 60
    deg_max = camera_fov_deg * (width / 2 - x_min) / width  # small value on the left corresponds to positive angle
    deg_min = camera_fov_deg * (width / 2 - x_max) / width
    tolerance = int(5 * angular_resolution)  # in paritular the valve is detected with offset
    left_index = mid_index + int(deg_min * angular_resolution) - tolerance
    right_index = mid_index + int(deg_max * angular_resolution) + tolerance
#    if verbose:
#        print('SubSelection', deg_min, deg_max, left_index, right_index, scan[left_index:right_index])

    tmp = [x if x > 0 else 100000 for x in scan]
    dist_mm = min(tmp[left_index:right_index])
    index = left_index + tmp[left_index:right_index].index(dist_mm)
    deg_100th = int(((index / angular_resolution) - 135) * 100)
    return deg_100th, dist_mm


class ArtifactDetector(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register("artf", "dropped", "debug_artf", "stdout")
        self.is_virtual = config['virtual_world']  # say different "dataset" (now red or all object detection)
        self.best = None
        self.best_count = 0
        self.best_img = None
        self.best_info = None
        self.best_scan = None
        self.best_depth = None
        self.verbose = False
        self.dump_dir = None  # optional debug ouput into directory
        self.scan = None  # should laster initialize super()
        self.depth = None  # more precise definiton of depth image
        self.width = None  # detect from incoming images

        self.gas_best = None
        self.gas_best_count = 0

    def handle_gas_artifact(self, data):
        print(self.time, 'Gas detected', data)
        if data:  # in virtual world only Boolean is used and transition is reported
            deg_100th, dist_mm = 0, 0  # first approximation without scan and entrance detection
            self.publish('artf', [GAS, deg_100th, dist_mm])

    def waitForImage(self):
        channel = ""
        while channel != "image":
            self.time, channel, data = self.listen()
            setattr(self, channel, data)

            # handling special artifact "gas", which is not requiring image
            if channel == "co2":
                if self.gas_best is None or self.gas_best < data:
                    self.gas_best = data
                    self.gas_best_count = 10
                    print(self.time, 'GAS CO2 value', self.gas_best)

                if self.gas_best > CO2_REPORT_LIMIT and self.gas_best_count == 0:
                    self.handle_gas_artifact(True)
                    self.gas_best = None

                if self.gas_best_count > 0:
                    self.gas_best_count -= 1
            elif channel == "wifiscan":
               for name, signal in data:
                    if name.startswith(PHONE_PREFIX) and signal > PHONE_SIGNAL_LIMIT:
                        deg_100th, dist_mm = 0, 0  # approximation
                        self.publish('artf', [PHONE, deg_100th, dist_mm])

        return self.time

    def stdout(self, *args, **kwargs):
        # maybe refactor to Node?
        output = StringIO()
        print(*args, file=output, **kwargs)
        contents = output.getvalue().strip()
        output.close()
        self.publish('stdout', contents)
        print(contents)

    def run(self):
        try:
            dropped = 0
            while True:
                now = self.publish("dropped", dropped)
                dropped = -1
                timestamp = now
                while timestamp <= now:
                    timestamp = self.waitForImage()
                    dropped += 1
                self.detect(self.image)
        except BusShutdownException:
            pass

    def detect(self, image):
        img = cv2.imdecode(np.fromstring(image, dtype=np.uint8), 1)
        if self.width is None:
            self.stdout('Image resolution', img.shape)
            self.width = img.shape[1]
        assert self.width == img.shape[1], (self.width, img.shape[1])
        phone_count, w, h, x_min, x_max = count_orange_blue(img)
        if phone_count >= 25:
            print(self.time, 'phone', phone_count)
            artf = PHONE
            deg_100th, dist_mm = 0, 500  # in front of the robot
            self.publish('artf', [artf, deg_100th, dist_mm])
            self.publish('debug_artf', image)  # JPEG
            if self.dump_dir is not None:
                filename = 'artf_%s_%d.jpg' % (artf, self.time.total_seconds())
                with open(os.path.join(self.dump_dir, filename), 'wb') as f:
                    f.write(image)
        rcount, w, h, x_min, x_max = count_red(img)
        yellow_used = False
        if True and rcount < 20:
            ycount, w, h, x_min, x_max = count_yellow(img)
            if ycount > YELLOW_THRESHOLD:
                yellow_used = True
                count = ycount
            else:
                count = rcount
        else:
            count = rcount

        if self.verbose and count >= 20:
            print(self.time, img.shape, count, w, h, x_min, x_max, w/h, count/(w*h))
        if self.best_count > 0:
            self.best_count -= 1
        if count > RED_THRESHOLD or (yellow_used and count > YELLOW_THRESHOLD):
            if self.best is None or count > self.best:
                self.best = count
                self.best_count = 10
                self.best_img = self.image
                self.best_info = w, h, x_min, x_max, (count == rcount), yellow_used  # RED used
                self.best_scan = self.scan
                self.best_depth = self.depth

        if self.best is not None and self.best_count == 0:
            w, h, x_min, x_max, red_used, yellow_used = self.best_info
            if red_used:
                # revise noisy points
                img = cv2.imdecode(np.fromstring(self.best_img, dtype=np.uint8), 1)
                rcount, w, h, x_min, x_max = count_red(img, filtered=True, stdout=self.stdout)
                if rcount == 0:
                    self.stdout('Invalid after filtering! orig count=', self.best)
                    # reset detector
                    self.best = None
                    self.best_count = 0
                    self.best_img = None
                    self.best_info = None
                    self.best_scan = None
                    self.best_depth = None
                    return
                self.stdout(rcount, w, h, x_min, x_max, w/h, count/(w*h))

            if red_used or yellow_used:
                if self.best_depth is not None:
                    global g_mask
                    g_mask = None
                    img = cv2.imdecode(np.fromstring(self.best_img, dtype=np.uint8), 1)
                    if red_used:
                        count_red(img)
                    else:
                        count_yellow(img)
                    assert g_mask is not None  # intermediate results
                    dist_mm = int(np.median(self.best_depth[g_mask]))
                    mask2 = np.abs(self.best_depth - dist_mm) < 200
                    mask = np.logical_and(g_mask, mask2)
                    # debug
                    #img2 = img.copy()
                    #img2[:] = (0, 0, 0)
                    #img2[g_mask] = (0, 0, 255)
                    #img2[mask] = (0, 255, 0)
                    #cv2.imwrite('artf.jpg', img2)

                    count, w, h, x_min, x_max = count_mask(mask)
                    FX = 462.1  # Focal length.
                    if (dist_mm > 10000 or count < RED_YELLOW_MIN_3D_THRESHOLD or # mix of infinity
                            (red_used and dist_mm * h/FX < 210) or                # robot
                            (yellow_used and dist_mm * h/FX > 850)):              # too high for survivor
                        self.stdout('Invalid distance, ignore, count=', self.best, count, dist_mm, dist_mm * h/FX)
                        # reset detector
                        self.best = None
                        self.best_count = 0
                        self.best_img = None
                        self.best_info = None
                        self.best_scan = None
                        self.best_depth = None
                        return

                    deg_100th = int(round(100 * 69.4 * (self.width/2 - (x_min + x_max)/2)/self.width))

#                    img2 = img.copy()
#                    img2[mask] = (0, 0, 255)
#                    img2[np.logical_not(mask)] = (0, 0, 0)
#                    cv2.imwrite('artfX.jpg', img2)
                else:
                    deg_100th, dist_mm = artf_in_scan(self.best_scan, self.width, x_min, x_max, verbose=True)
            else:
                deg_100th, dist_mm = 0, 500  # in front of the robot

            if red_used:
                self.stdout('h/w', h/w)
#                if self.best < 1000:
#                    artf = DRILL  # VALVE - hack for simple02
#                elif h/w > 2.4:
#                    artf = EXTINGUISHER
#                elif h/w > 1.0:
#                    artf = BACKPACK
#                else:
#                    artf = TOOLBOX
#                if self.best < 1000:
#                    artf = DRILL  # VALVE - hack for simple02, fallback for empty image
#                if h/w > 2:
#                    artf = EXTINGUISHER
                artf = BACKPACK
            elif yellow_used:
                if self.best > YELLOW_MAX_THRESHOLD:
                    # too much yellow - barrel or yellow machine
                    self.best = None
                    self.best_count = 0
                    self.best_img = None
                    self.best_info = None
                    self.best_scan = None
                    self.best_depth = None
                    return
                artf = RESCUE_RANDY  # used to be RADIO
            self.stdout(self.time, 'Relative position:', self.best, deg_100th, dist_mm, artf)

            dx_mm, dy_mm = 0, 0  # relative offset to current robot position
            # TODO if VALVE -> find it in scan
            self.publish('artf', [artf, deg_100th, dist_mm])
            self.publish('debug_artf', self.best_img)
            if self.dump_dir is not None:
                filename = 'artf_%s_%d.jpg' % (artf, self.time.total_seconds())
                with open(os.path.join(self.dump_dir, filename), 'wb') as f:
                    f.write(self.best_img)
                if self.best_depth is not None:
                    filename = 'artf_%s_%d.npz' % (artf, self.time.total_seconds())
                    np.savez_compressed(os.path.join(self.dump_dir, filename), depth=self.best_depth)

            # reset detector
            self.best = None
            self.best_count = 0
            self.best_img = None
            self.best_info = None
            self.best_scan = None
            self.best_depth = None


def debug2dir(filename, out_dir):
    from osgar.logger import LogReader, lookup_stream_names
    from osgar.lib.serialize import deserialize

    names = lookup_stream_names(filename)
    assert 'detector.debug_artf' in names, names
    assert 'detector.artf' in names, names
    assert 'rosmsg.sim_time_sec' in names, names
    image_id = names.index('detector.debug_artf') + 1
    artf_id = names.index('detector.artf') + 1
    sim_sec_id = names.index('rosmsg.sim_time_sec') + 1
    sim_time_sec = None
    image = None
    artf = None
    for dt, channel, data in LogReader(filename, only_stream_id=[image_id, artf_id, sim_sec_id]):
        data = deserialize(data)
        if channel == sim_sec_id:
            sim_time_sec = data
        elif channel == image_id:
            image = data
            assert artf is not None
            time_sec = sim_time_sec if sim_time_sec is not None else int(dt.total_seconds())
            name = os.path.basename(filename)[:-4] + '-' + artf[0] + '-' + str(time_sec) + '.jpg'
            print(name)
            with open(os.path.join(out_dir, name), 'wb') as f:
                f.write(image)
        elif channel == artf_id:
            artf = data


if __name__ == '__main__':
    from unittest.mock import MagicMock
    from queue import Queue
    import argparse
    import datetime
    import sys
    from osgar.bus import Bus

    parser = argparse.ArgumentParser(description='Run artifact detection and classification for given JPEG image')
    parser.add_argument('filename', help='JPEG filename')
    parser.add_argument('--debug2dir', help='dump clasified debug images into directory')
    parser.add_argument('--depth', help='filename of depth image for tested together with JPEG image')
    parser.add_argument('-v', '--verbose', help='verbose mode', action='store_true')
    args = parser.parse_args()

    if args.debug2dir is not None:
        debug2dir(args.filename, args.debug2dir)
        sys.exit()

    with open(args.filename.replace('.npz', '.jpg'), 'rb') as f:
        jpeg_data = f.read()

    config = {'virtual_world': True}  # for now
    logger = MagicMock()
    logger.register = MagicMock(return_value=1)
    def counter():
        start = datetime.datetime.utcnow()
        while True:
            dt = datetime.datetime.utcnow() - start
            yield dt
    logger.write = MagicMock(side_effect=counter())
    bus = Bus(logger)
    detector = ArtifactDetector(config, bus.handle('detector'))
    detector.verbose = args.verbose
    tester = bus.handle('tester')
    tester.register('scan', 'image', 'tick', 'depth')
    bus.connect('tester.scan', 'detector.scan')
    bus.connect('tester.depth', 'detector.depth')
    bus.connect('tester.image', 'detector.image')
    bus.connect('detector.artf', 'tester.artf')
    bus.connect('tester.tick', 'tester.tick')
    bus.connect('detector.dropped', 'tester.dropped')
    tester.publish('scan', [2000]*270)  # pretend that everything is at 2 meters
    if args.depth is not None or args.filename.endswith('.npz'):
        filename = args.depth if args.depth is not None else args.filename
        with np.load(filename) as f:
            depth = f['depth']
            tester.publish('depth', depth)
    detector.start()
    for i in range(10 + 1):  # workaround for local minima
        a = tester.listen()
#        print(i, a)
        tester.sleep(0.01)
        tester.publish('image', jpeg_data)
    detector.request_stop()
    detector.join()
    tester.publish('tick', None)
    a = tester.listen()
    print(a)

# vim: expandtab sw=4 ts=4



import unittest
import math
from unittest.mock import MagicMock

from osgar.drivers.kloubak import (compute_desired_erpm, compute_desired_angle,
        WHEEL_DISTANCE, compute_rear, CENTER_AXLE_DISTANCE, RobotKloubak,
        get_downdrop_bumpers, MAX_JOINT_ANGLE, CAN_ID_ENCODERS, ENC_SCALE)


class KloubakTest(unittest.TestCase):

    def test_compute_desired_erpm(self):
        # go straight 1m/s
        left, right = compute_desired_erpm(1.0, 0.0)
        self.assertEqual(left, right)
        self.assertEqual(left, 694)  # 917 for 25" wheels and 694 for 33"

        wd = WHEEL_DISTANCE

        # left wheel should stay and only right one turn, say 180deg/sec
        left, right = compute_desired_erpm(math.pi * wd / 2, math.pi)
        self.assertEqual(left, 0)
        self.assertEqual(right, 1082)  # 1428 for 25" wheels and 1082 for 33"

        # rotate left with left moving the half the speed of right
        # left should be "wd", right "2*wd"
        left, right = compute_desired_erpm((wd/2 + wd)/2, 0.5)
        self.assertAlmostEqual(left, right / 2, 0)

        # backup 0.5m/s
        left, right = compute_desired_erpm(-0.5, 0.0)
        self.assertEqual(left, right)
        self.assertEqual(left, -347)  # -458 for 25" wheels and -347 for 33"

    def test_compute_rear(self):
        speed, angular = compute_rear(1.0, 0.0, 0.0)
        self.assertAlmostEqual(speed, 1.0)
        self.assertAlmostEqual(angular, 0.0)

        # turn in place
        speed, angular = compute_rear(0.0, 0.1, math.pi/2)
        self.assertAlmostEqual(speed, 0.1 * CENTER_AXLE_DISTANCE)
        self.assertAlmostEqual(angular, 0.0)

    def test_rear_drive(self):
        # go straight
        self.assertAlmostEqual(0.0, compute_desired_angle(1.0, 0.0))

        # easy case when the joint has 60 degrees and radius is CENTER_AXLE_DISTANCE
        self.assertAlmostEqual(0.9272952,
                compute_desired_angle(CENTER_AXLE_DISTANCE, 0.5))

        # it should be symmetrical
        self.assertAlmostEqual(-0.9272952,
                compute_desired_angle(CENTER_AXLE_DISTANCE, -0.5))

        # perpendicular to the left, radius defined by length of joint to wheel center
#        self.assertAlmostEqual(math.radians(90),
#                compute_desired_angle(CENTER_AXLE_DISTANCE * math.radians(90), math.radians(90)))

        # the same on the other side
#        self.assertAlmostEqual(math.radians(-90),
#                compute_desired_angle(CENTER_AXLE_DISTANCE * math.radians(90), math.radians(-90)))

        # failing case "ValueError: math domain error"
        compute_desired_angle(0.1, -0.007853981633974483)

        # failing case "ValueError: math domain error" after fix - limits needed
        compute_desired_angle(0.1, 0.3063052837250049)

        # another case "ValueError: math domain error"
        angle = compute_desired_angle(0.1, -1.0637781790905438)
        self.assertAlmostEqual(angle, -1.1868238913561442)

    def test_invalid_can_message(self):
        # this message killed Kloubak K2 on DARPA SubT Tunnel Circuit, Day 2
        config = {}
        bus = MagicMock()
        k2 = RobotKloubak(config, bus)
        # this used to assert:
        #      assert len(data) == 1, len(data)
        #   AssertionError: 8
        k2.update_buttons(b'\x00\x00\x04\x1f\x00\x1f\x00\x7f')
        self.assertEqual(k2.can_errors, 1)

        k2.update_encoders(msg_id=0x92, data=b'\x97')

    def test_partial_axis_encoders_update(self):
        config = {
            'num_axis': 3
        }
        bus = MagicMock()
        k3 = RobotKloubak(config, bus)
        self.assertIsNone(k3.partial_axis_encoders_update((121, 2, 2280, 2486)))

        # epoch change, but incomplete data -> do not return anything
        self.assertIsNone(k3.partial_axis_encoders_update((122, 1, 2183, 3786)))

        # now proper update
        self.assertIsNone(k3.partial_axis_encoders_update((122, 2, 2280, 2486)))
        self.assertIsNone(k3.partial_axis_encoders_update((122, 3, 2164, 2486)))

        # epoch change
        self.assertEqual(k3.partial_axis_encoders_update((123, 1, 2183, 3786)),
                         [2183, 3786, 2280, 2486, 2164, 2486])

    def test_downdrops(self):
        self.assertEqual(get_downdrop_bumpers([500, 500]), [False, False])
        self.assertEqual(get_downdrop_bumpers([1000, 200]), [True, True])

    def test_compute_desired_angle_division_by_zero(self):
        self.assertAlmostEqual(compute_desired_angle(desired_speed=0.0, desired_angular_speed=0.1),
                               MAX_JOINT_ANGLE)

        self.assertAlmostEqual(compute_desired_angle(desired_speed=0.0, desired_angular_speed=-20.1),
                               -MAX_JOINT_ANGLE)

    def test_publish_encoders(self):
        config = {
            'num_axis': 3
        }
        bus = MagicMock()
        k3 = RobotKloubak(config, bus)

        flags = 0
        k3.process_packet([CAN_ID_ENCODERS, bytes([13, 1, 0, 0, 0, 0]), flags])
        bus.publish.assert_not_called()
        k3.process_packet([CAN_ID_ENCODERS, bytes([13, 2, 0, 0, 0, 0]), flags])
        bus.publish.assert_not_called()
        k3.process_packet([CAN_ID_ENCODERS, bytes([13, 3, 0, 0, 0, 0]), flags])
        bus.publish.assert_not_called()

        k3.process_packet([CAN_ID_ENCODERS, bytes([14, 1, 0, 1, 0, 2]), flags])
        bus.publish.assert_called()

    def test_kloubak_heading(self):
        config = {
            'num_axis': 2,
            "wheel_distance": 0.496
        }
        bus = MagicMock()
        k2 = RobotKloubak(config, bus)

        # forward moving, use joint angle for correction
        k2.encoders = [100, 100, 100, 100 + WHEEL_DISTANCE/ENC_SCALE]
        k2.last_pose_encoders = [100, 100, 100, 100]
        k2.last_joint_angle = 1
        k2.prev_joints_angle = 0
        k2.desired_speed = 1
        k2.update_pose()
        x, y, heading = k2.pose
        self.assertAlmostEqual(heading, 0)

        # backward moving, do not use joint angle for correction
        k2.encoders = [100, 100 + WHEEL_DISTANCE / ENC_SCALE, 100, 100]
        k2.last_pose_encoders = [100, 100, 100, 100]
        k2.last_joint_angle = 1
        k2.prev_joints_angle = 0
        k2.desired_speed = -1
        k2.update_pose()
        x, y, heading = k2.pose
        self.assertAlmostEqual(heading, -1)

# vim: expandtab sw=4 ts=4



import unittest
from unittest.mock import MagicMock, call

from subt.artf_gas import ArtifactGasDetector
from subt.artifacts import GAS


class ArtifactGasDetectorTest(unittest.TestCase):

    def test_report(self):
        bus = bus=MagicMock()
        detector = ArtifactGasDetector(bus=bus, config={})
        detector.on_gas_detected(False)
        bus.publish.assert_not_called()

        detector.on_gas_detected(True)
        bus.publish.assert_called()
        bus.publish.assert_has_calls([call('artf', [GAS, [0, 0, 0]])])

# vim: expandtab sw=4 ts=4



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="osgar",
    version="0.3.0",
    install_requires=[
          'msgpack>=1.0.0',
      ],
    extras_require={
        'tools': ['opencv-python>=3', 'Pygame'],
    },
    author="Robotika.cz",
    author_email="osgar@robotika.cz",
    description="Open Source Garden/Generic Autonomous Robot / Python library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/robotika/osgar",
    python_requires=">=3.6",
    packages=['osgar', 'osgar.drivers', 'osgar.lib', 'osgar.tools', 'subt'],
    package_data={
        '': ['config/*.json'],
    },
    entry_points={
        'console_scripts': [
            'lidarview = osgar.tools.lidarview:main [tools]',
            'log2video = osgar.tools.log2video:main [tools]',
            'log2pcap = osgar.tools.log2pcap:main [tools]',
            'logger = osgar.logger:main',
            'record = osgar.record:main',
            'replay = osgar.replay:main',
            'subt = subt.main:main',
            'allsync = subt.tools.allsync:main',
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)

# vim: expandtab sw=4 ts=4



"""
  Internal bus for communication among modules
"""
import inspect
import threading  # Event

from queue import Queue
from datetime import timedelta
from collections import deque

from osgar.lib.serialize import serialize, deserialize, compress


# restrict replay time from given input
ASSERT_QUEUE_DELAY = timedelta(seconds=.1)

def almost_equal(data, ref_data):
    if isinstance(data, float) and isinstance(ref_data, float):
        if data == ref_data:
            return True  # handling float('inf')
        return abs(data - ref_data) < 1e-6
    if isinstance(data, list) and isinstance(ref_data, list):
        if len(data) != len(ref_data):
            return False
        for a, b in zip(data, ref_data):
            if not(almost_equal(a, b)):
                return False
        return True
    # default - use exact match
    return data == ref_data


class BusShutdownException(Exception):
    pass


class Bus:
    def __init__(self, logger):
        self.logger = logger
        self.handles = {}

    def handle(self, name):
        return self.handles.setdefault(name, _BusHandler(self.logger, name))

    def connect(self, sender, receiver, modules=None):
        sender, output = sender.split('.')
        receiver, input = receiver.split('.')
        self.handles[sender].connect(output, self.handles[receiver], input, modules)

    def delays(self):
        for name, handle in self.handles.items():
            if handle.max_delay > ASSERT_QUEUE_DELAY:
                yield name, handle.max_delay, handle.max_delay_timestamp


class _BusHandler:
    def __init__(self, logger, name):
        self.logger = logger
        self.queue = Queue()
        self.name = name
        self.out = {}
        self.slots = {}
        self.stream_id = {}
        self.finished = threading.Event()
        self.no_output = set()
        self.compressed_output = set()
        self._time = None
        self.max_delay = timedelta()
        self.max_delay_timestamp = timedelta()

    def register(self, *outputs):
        for name_and_type in outputs:
            o = name_and_type.split(':')[0]  # keep only prefixes
            if o in self.stream_id:
                continue
            idx = self.logger.register(f'{self.name}.{o}')
            self.stream_id[o] = idx
            if name_and_type.endswith(':null'):
                self.no_output.add(idx)
            if name_and_type.endswith(':gz'):
                self.compressed_output.add(idx)
            self.out[o] = []
            self.slots[o] = []

    def connect(self, output, receiver, input, modules):
        if output not in self.out:
            #raise RuntimeError(f"'{self.name}' does not have '{output}' as output.")
            print(f"'{self.name}' does not have '{output}' as output.")
            return  # TODO parametrize by -F (force) or --ignore-missing_inputs??
        if input.startswith('slot_'):
            assert modules is not None
            assert receiver.name in modules
            self.slots[output].append(getattr(modules[receiver.name], input))
        else:
            self.out[output].append((receiver.queue, input))

    def publish(self, channel, data):
        with self.logger.lock:
            stream_id = self.stream_id[channel]  # local maping of indexes
            if stream_id in self.no_output:
                to_write = serialize(None)  # i.e. at least there will be timestamp record
            else:
                to_write = serialize(data)
                if stream_id in self.compressed_output:
                    to_write = compress(to_write)
            timestamp = self.logger.write(stream_id, to_write)

            if self._time is not None:
                delay = timestamp - self._time
                if delay > self.max_delay:
                    self.max_delay = delay
                    self.max_delay_timestamp = timestamp
                if delay > ASSERT_QUEUE_DELAY:
                    caller = inspect.currentframe().f_back.f_code.co_name
                    self.report_error(delay=delay.total_seconds(), channel=channel, caller=caller)

            for queue, input_channel in self.out[channel]:
                queue.put((timestamp, input_channel, data))
            for slot in self.slots.get(channel, []):
                slot(timestamp, data)
        return timestamp

    def listen(self):
        packet = self.queue.get()
        if packet is None:
            raise BusShutdownException()
        self._time, channel, data = packet
        return self._time, channel, data

    def sleep(self, secs):
        self.finished.wait(secs)

    def is_alive(self):
        return not self.finished.is_set()

    def shutdown(self):
        self.finished.set()
        self.queue.put(None)

    def report_error(self, **err):
        with self.logger.lock:
            data = dict(name=self.name)
            err.pop('name', None)
            data.update(err)
            self.logger.write(0, bytes(str(data), encoding='ascii'))


class LogBusHandler:
    def __init__(self, log, inputs, outputs):
        self.reader = log
        self.inputs = inputs
        self.outputs = outputs
        self.buffer_queue = deque()
        self.max_delay = timedelta()
        self.max_delay_timestamp = timedelta()
        self.finished = threading.Event()

    def register(self, *outputs):
        for o in outputs:
            # ignore :gz and :null modifiers
            assert o.split(':')[0] in self.outputs.values(), (o, self.outputs)

    def listen(self):
        while True:
            if self.finished.is_set():
                raise BusShutdownException
            if len(self.buffer_queue) == 0:
                dt, stream_id, bytes_data = next(self.reader)
            else:
                dt, stream_id, bytes_data = self.buffer_queue.popleft()
            if stream_id not in self.inputs:
                error = f"unexpected input (stream {stream_id}, known inputs {list(self.inputs.keys())})"
                raise RuntimeError(error)
            channel = self.inputs[stream_id]
            data = deserialize(bytes_data)
            if channel.startswith("slot_"):
                getattr(self.node, channel)(dt, data)
            else:
                break
        return dt, channel, data

    def publish(self, channel, data):
        assert channel in self.outputs.values(), (channel, tuple(self.outputs.values()))
        dt, stream_id, bytes_data = next(self.reader)
        while stream_id not in self.outputs:
            input_name = self.inputs[stream_id]
            if input_name.startswith("slot_"):
                data = deserialize(bytes_data)
                getattr(self.node, input_name)(dt, data)
            else:
                self.buffer_queue.append((dt, stream_id, bytes_data))
            dt, stream_id, bytes_data = next(self.reader)
        assert channel == self.outputs[stream_id], (channel, self.outputs[stream_id], dt)  # wrong channel
        if len(self.buffer_queue) > 0:
            delay = dt - self.buffer_queue[0][0]
            if delay > self.max_delay:
                self.max_delay = delay
                self.max_delay_timestamp = dt
            if delay > ASSERT_QUEUE_DELAY:
                print(dt, "maximum delay overshot:", delay)
        ref_data = deserialize(bytes_data)
        assert almost_equal(data, ref_data), (data, ref_data, dt)
        return dt

    def sleep(self, secs):
        pass

    def shutdown(self):
        raise SystemExit()

    def report_error(self, err):
        print(self.time, err)


class LogBusHandlerInputsOnly:
    def __init__(self, log, inputs):
        self.reader = log
        self.inputs = inputs
        self.time = timedelta(0)

    def register(self, *outputs):
        pass

    def listen(self):
        dt, stream_id, bytes_data = next(self.reader)
        self.time = dt
        channel = self.inputs[stream_id]
        data = deserialize(bytes_data)
        if channel.startswith("slot_"):
            channel = channel[len('slot_'):]  # workaround for non-slot run
        return dt, channel, data

    def publish(self, channel, data):
        return self.time

    def sleep(self, secs):
        pass

    def shutdown(self):
        pass

    def report_error(self, err):
        print(self.time, err)


if __name__ == "__main__":
    pass

# vim: expandtab sw=4 ts=4



import math

from osgar.lib.loop import LoopDetector
from osgar.lib.quaternion import euler_to_quaternion
from osgar.lib.unittest import TestCase

class LoopDetectionTest(TestCase):
    def test_perfect_loop(self):
        forward = euler_to_quaternion(0, 0, 0)
        trajectory = [((0, 0, 0), forward),
                      ((2, 0, 0), forward),
                      ((2, 2, 0), forward),
                      ((0, 2, 0), forward),
                      ((0, 0, 0), forward)]
        detector = LoopDetector(min_loop_length=5)
        detector.add_all(trajectory)
        loop = detector.loop()
        self.assertEqual(loop, trajectory)


    def test_loop_with_prefix(self):
        forward = euler_to_quaternion(0, 0, 0)
        trajectory = [((-1, 0, 0), forward),
                      ((0, 0, 0), forward),
                      ((2, 0, 0), forward),
                      ((2, 2, 0), forward),
                      ((0, 2, 0), forward),
                      ((0, 0, 0), forward)]
        detector = LoopDetector(min_loop_length=5)
        detector.add_all(trajectory)
        loop = detector.loop()
        self.assertEqual(loop, trajectory[1:])


    def test_loop_with_small_rotation(self):
        forward = euler_to_quaternion(0, 0, 0)
        almost_forward = euler_to_quaternion(math.radians(10), 0, 0)
        trajectory = [((0, 0, 0), forward),
                      ((2, 0, 0), forward),
                      ((2, 2, 0), forward),
                      ((0, 2, 0), forward),
                      ((0, 0, 0), almost_forward)]
        detector = LoopDetector(min_loop_length=5)
        detector.add_all(trajectory)
        loop = detector.loop()
        self.assertEqual(loop, trajectory)


    def test_loop_with_small_drift(self):
        forward = euler_to_quaternion(0, 0, 0)
        trajectory = [((0, 0, 0), forward),
                      ((2, 0, 0), forward),
                      ((2, 2, 0), forward),
                      ((0, 2, 0), forward),
                      ((0.1, 0.1, 0.1), forward)]
        detector = LoopDetector(min_loop_length=5)
        detector.add_all(trajectory)
        loop = detector.loop()
        self.assertEqual(loop, trajectory)


    def test_granularity(self):
        forward = euler_to_quaternion(0, 0, 0)
        trajectory = [((0, 0, 0), forward),
                      ((1, 0, 0), forward),
                      ((2, 0, 0), forward),
                      ((2, 1, 0), forward),
                      ((2, 2, 0), forward),
                      ((1, 2, 0), forward),
                      ((0, 2, 0), forward),
                      ((0, 1, 0), forward),
                      ((0, 0, 0), forward)]
        detector = LoopDetector(min_loop_length=5, granularity=2.0)
        detector.add_all(trajectory)
        loop = detector.loop()
        self.assertEqual(loop, trajectory[::2])


    def test_not_a_loop_too_far_horizontally(self):
        forward = euler_to_quaternion(0, 0, 0)
        trajectory = [((0, 0, 0), forward),
                      ((2, 0, 0), forward),
                      ((2, 2, 0), forward),
                      ((0, 2, 0), forward)]
        detector = LoopDetector(min_loop_length=5)
        detector.add_all(trajectory)
        loop = detector.loop()
        self.assertIsNone(loop)


    def test_not_a_loop_too_far_vertically(self):
        forward = euler_to_quaternion(0, 0, 0)
        trajectory = [((0, 0, 0), forward),
                      ((2, 0, 0), forward),
                      ((2, 2, 0), forward),
                      ((0, 2, 0), forward),
                      ((0, 0, 1), forward)]
        detector = LoopDetector(min_loop_length=5)
        detector.add_all(trajectory)
        loop = detector.loop()
        self.assertIsNone(loop)


    def test_not_a_loop_different_direction(self):
        forward = euler_to_quaternion(0, 0, 0)
        left = euler_to_quaternion(math.pi/2, 0, 0)
        trajectory = [((0, 0, 0), forward),
                      ((2, 0, 0), forward),
                      ((2, 2, 0), forward),
                      ((0, 2, 0), forward),
                      ((0, 0, 0), left)]
        detector = LoopDetector(min_loop_length=5)
        detector.add_all(trajectory)
        loop = detector.loop()
        self.assertIsNone(loop)


    def test_not_a_loop_too_short(self):
        forward = euler_to_quaternion(0, 0, 0)
        trajectory = [((0, 0, 0), forward),
                      ((1, 0, 0), forward),
                      ((1, 1, 0), forward),
                      ((0, 1, 0), forward),
                      ((0, 0, 0), forward)]
        detector = LoopDetector(min_loop_length=5)
        detector.add_all(trajectory)
        loop = detector.loop()
        self.assertIsNone(loop)


    def test_not_a_loop_too_long(self):
        forward = euler_to_quaternion(0, 0, 0)
        trajectory = [((0, 0, 0), forward),
                      ((100, 0, 0), forward),
                      ((100, 100, 0), forward),
                      ((0, 100, 0), forward),
                      ((0, 0, 0), forward)]
        detector = LoopDetector(min_loop_length=5)
        detector.add_all(trajectory)
        loop = detector.loop()
        self.assertIsNone(loop)


if __name__ == '__main__':
    import unittest
    unittest.main()



import unittest
import os.path
from subt.dummy_darpa_server import GameLogic, dist3d


class DummyDarpaServerTest(unittest.TestCase):

    def test_game_logic(self):
        filename = os.path.join(os.path.dirname(__file__), 'config', 'darpa-artf.csv')
        # content
        """
artf, x, y, z
Backpack, 2, -21.2, 0.5
Cell Phone, 5, 5, 5
        """
        game = GameLogic(filename)
        self.assertEqual(game.score, 0)

        self.assertTrue(game.report_artf('Cell Phone', (5, 5, 5)))  # exact match
        self.assertEqual(game.score, 1)

        self.assertFalse(game.report_artf('Cell Phone', (5, 5, 5)))  # already reported

    def test_dist3d(self):
        self.assertAlmostEqual(dist3d((0, 0, 0), (1, 0, 0)), 1.0)
        self.assertAlmostEqual(dist3d((1, 2, 3), (1+3, 2+4, 3)), 5.0)


# vim: expandtab sw=4 ts=4




from threading import Thread


class Timer:
    def __init__(self, config, bus):
        self.bus = bus
        bus.register('tick')
        self.sleep = config['sleep']
        self.thread = Thread(target=self.run)

    def start(self):
        self.thread.start()

    def join(self, timeout=None):
        self.thread.join(timeout=timeout)

    def run(self):
        while self.bus.is_alive():
            self.bus.publish('tick', None)
            self.bus.sleep(self.sleep)

    def request_stop(self):
        self.bus.shutdown()



import unittest
from unittest.mock import MagicMock
from datetime import timedelta

from osgar.drivers.spider import Spider, CAN_packet
from osgar.bus import Bus

class SpiderTest(unittest.TestCase):

    def test_split_buffer(self):
        self.assertEqual(Spider.split_buffer(b''), (b'', b''))
        self.assertEqual(Spider.split_buffer(b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xfe\x10'), (b'', b'\xfe\x10'))
        data = b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xfeW\xfe0@h\x9e\x01i\x01\xf7\x01\x18\x00'
        self.assertEqual(Spider.split_buffer(data), (b'\xfe0@h\x9e\x01i\x01\xf7\x01\x18\x00', b'\xfeW'))

        data = b'0@h\x9e\x01i\x01\xf7\x01\x18\x00'
        self.assertEqual(Spider.split_buffer(data), (b'h\x9e\x01i\x01\xf7\x01\x18\x00', b'0@'))

    def test_can_packet(self):
        self.assertEqual(CAN_packet(0x400, [0, 0]), b'\x80\x02\x00\x00')

    def test_uninitialized_can_bridge(self):
        bus = MagicMock()
        spider = Spider(config={'stream_id_in':1, 'stream_id_out':2}, bus=bus)
        spider.send((0, 0))
#        bus.write.assert_called_once_with(0, 'ERROR: CAN bridge not initialized yet! [(0, 0)]')

    def test_publish_status(self):
        logger=MagicMock()
        logger.write = MagicMock(return_value=timedelta(seconds=135))
        bus = Bus(logger=logger)
        tester = bus.handle('tester')
        tester.register('raw')
        spider = Spider(config={}, bus=bus.handle('spider'))
        bus.connect('tester.raw', 'spider.raw')
        bus.connect('spider.status', 'tester.status')

        spider.can_bridge_initialized = True  # skip initialization
        self.assertEqual(CAN_packet(0x200, [0, 0x80]), b'@\x02\x00\x80')
        tester.publish('raw', b'@\x02\x00\x80')
        spider.start()
        dt, stream, data = tester.listen()
        spider.request_stop()
        spider.join()
        self.assertEqual(dt, timedelta(seconds=135))
        self.assertEqual(stream, 'status')
        self.assertEqual(data, ([0x8000, None]))

# vim: expandtab sw=4 ts=4



"""
  Wrapper & timestamper of input serial byte stream
"""

import serial
from threading import Thread

from osgar.logger import LogWriter
from osgar.bus import BusShutdownException


class LogSerial:
    def __init__(self, config, bus):
        bus.register('raw')
        self.input_thread = Thread(target=self.run_input, daemon=True)
        self.output_thread = Thread(target=self.run_output, daemon=True)

        if 'port' in config:
            if config.get('rtscts'):
                self.com = serial.Serial(config['port'], config['speed'], rtscts=True)
                self.com.setRTS()
            else:
                self.com = serial.Serial(config['port'], config['speed'])
            self.com.timeout = config.get('timeout', 0.01)  # default expects updates < 100Hz
            if config.get('reset'):
                self.com.setDTR(0)
        else:
            self.com = None
        self.bus = bus

        self.buf = b''

    def start(self):
        self.input_thread.start()
        self.output_thread.start()

    def join(self, timeout=None):
        self.input_thread.join(timeout=timeout)
        self.output_thread.join(timeout=timeout)

    def run_input(self):
        while self.bus.is_alive():
            data = self.com.read(1024)
            if len(data) > 0:
                self.bus.publish('raw', data)

    def slot_raw(self, timestamp, data):
        self.com.write(data)

    def run_output(self):
        try:
            while True:
                dt, __, data = self.bus.listen()
                self.slot_raw(dt, data)
        except BusShutdownException:
            pass

    def request_stop(self):
        self.bus.shutdown()


if __name__ == "__main__":
    import time
    from osgar.bus import Bus

    config = { 'port': 'COM5', 'speed': 4800 }
    with LogWriter(prefix='test-') as log:
        bus = Bus(log)
        device = LogSerial(config, bus=bus.handle('serial'))
        device.start()
        time.sleep(2)
        device.request_stop()
        device.join()

# vim: expandtab sw=4 ts=4





def create_load_tests(filepath):
    from pathlib import Path
    start_dir = Path(filepath).parent
    def load_tests(loader, standard_tests, pattern):
        pattern = pattern or "test_*.py"
        package_tests = loader.discover(start_dir=start_dir, pattern=pattern)
        standard_tests.addTests(package_tests)
        return standard_tests
    return load_tests


load_tests = create_load_tests(__file__)



"""
  LoRa - Long Range Radio

  This module handles communication with LoRa devices and automatic re-transmission
  of received messages so that all robots work as re-transmission stations.
  The messages have limited size 40 bytes.

  At the moment driver supports only transmission of "pose2d" messages so the control
  center (and all neighbor robots) know about each other.

  Current protocol uses numbers for DeviceID (1, 2, ... 5) for source identification
  and '|' for data separation. The whole packet is terminated by '\n'. The DeviceID
  is added in front. The message history is collected in reverse order.

  The transmitted messages are echo back with DeviceID prefix. The DeviceID is hardcoded
  in the device firmware (i.e. not controlled by OSGAR). The identification of connected
  device is implemented by transmitting pseudo-random message and parsing its echo.

  Example:
     b'2|5|4|hello' - original message "hello" was sent by device ID=4 and retransmitted
                      by device ID=5 and retransmitted again by device ID=2

  I/O:
  This driver can now parse commands for particular device in the form:
    <device_id>:<cmd>:<hash>
  The <device_id> has to match in order to publish <cmd>. <hash> is used only for distinction
  of messages, and the plan is to use seconds since control center start.
  Broadcast command for all robots is realized by <device_id> = 0.

"""
from datetime import timedelta
from ast import literal_eval

from osgar.node import Node
from osgar.bus import BusShutdownException

ALIVE_MESSAGE = b'alive'
ALLOWED_DEVICE_IDS = [1, 2, 3, 4, 5, 6]


def parse_lora_packet(packet):
    """
       Parse block of binary data. The block starts with integer
       numbers interlaced by '|'. List of transmitter IDs is returned
       with the actual payload data. In the case of missing transmitted
       history header None is returned.
    """
    s = packet.split(b'|')
    try:
        addr, data = [int(x) for x in s[:-1]], s[-1].strip()
    except ValueError:
        return None, packet

    if len(addr) < 1 or not min([a in ALLOWED_DEVICE_IDS for a in addr]):
        return None, packet
    return addr, data


def split_lora_buffer(buf):
    if b'\n' not in buf:
        return buf, b''
    i = buf.index(b'\n')
    return buf[i+1:], buf[:i+1]


def parse_my_cmd(my_id, data):
    s = data.split(b':')
    try:
        destination_id = int(s[0])
        if destination_id == my_id or destination_id == 0:
            return s[1]
    except:
        pass
    return None


def draw_lora_positions(arr):
    """
    Draw positions of tripples:
        (time, ID, (x. y, heading))
    """
    import matplotlib.pyplot as plt
    t = [a[0] for a in arr]
    robot_ids = sorted(list(set([a[1] for a in arr])))
    print('Robot IDs', robot_ids)

    for robot_id in robot_ids:
        x = [a[2][0]/1000.0 for a in arr if a[1] == robot_id]
        y = [a[2][1]/1000.0 for a in arr if a[1] == robot_id]
        line = plt.plot(x, y, '-o', linewidth=2, label='Robot #%d' % robot_id)

#    plt.xlabel('time (s)')
    plt.axes().set_aspect('equal', 'datalim')
    plt.legend()
    plt.show()


class LoRa(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register('raw', 'cmd', 'robot_status', 'artf', 'artf_xyz')
        self.device_id = config.get('device_id')  # None for "autodetect"
        self.init_sleep = config.get('sleep')  # needed for Windows on start
        self.min_transmit_dt = timedelta(seconds=10)  # TODO config?
        self.buf = b''
        self.raw = b''  # should be defined by Node
        self.last_transmit = None
        self.recent_packets = []
        self.verbose = config.get('verbose', False)
        self.debug_robot_poses = []

    def send_data(self, data):
        self.last_transmit = self.publish('raw', data + b'\n')
        return self.last_transmit

    def on_radio(self, data):
        src, packet = data
        name = src.decode('ascii')
        if packet.startswith(b'['):
            arr = literal_eval(packet.decode('ascii'))
            if len(arr) == 3:
                # position
                pass
            elif len(arr) == 4:
                # artifact
                self.publish('artf_xyz', [arr])  # publish also standard "list" of detected artifacts
            else:
                assert False, arr  # unexpected size/type

    def update(self):
        self.recent_packets = []
        channel = super().update()  # define self.time
        if channel == 'raw':
            if self.verbose:
                print(self.time, "update", channel, self.raw)

            self.buf, packet = split_lora_buffer(self.buf + self.raw)
            while len(packet) > 0:
                self.recent_packets.append(packet)
                addr, data = parse_lora_packet(packet)
                if addr is not None and data.startswith(b'['):
                    arr = literal_eval(data.decode('ascii'))
                    if len(arr) == 3:
                        pose2d = arr
                        self.publish('robot_status', [addr[-1], pose2d, b'running'])  # TODO read it from robot
                    else:
                        assert len(arr) == 4, arr  # artifact, x, y z
                        self.publish('artf', [addr[-1], arr])
                    if self.verbose:
                        self.debug_robot_poses.append((self.time.total_seconds(), addr[-1], pose))
                cmd = parse_my_cmd(self.device_id, data)
                if cmd is not None:
                    self.publish('cmd', cmd)
                if addr is not None and self.device_id is not None and self.device_id not in addr:
                    if self.verbose:
                        print('re-transmit')
                    self.send_data(packet.strip())  # re-transmit data from other robots
                self.buf, packet = split_lora_buffer(self.buf)
        elif channel == 'radio':
            self.on_radio(self.radio)
        elif channel == 'pose2d':
            if self.last_transmit is None or self.time > self.last_transmit + self.min_transmit_dt:
                self.send_data(bytes(str(self.pose2d), encoding='ascii'))
        elif channel == 'cmd':
            assert len(self.cmd) == 2, self.cmd
            self.send_data(b'%d:%s:%d' % (self.cmd[0], self.cmd[1], int(self.time.total_seconds())))
        elif channel == 'artf':
            # send data as they are, ignore transmit time, ignore transmit failure
            for artf_item in self.artf:
                self.send_data(bytes(str(artf_item), encoding='ascii'))
        else:
            assert False, channel  # not supported
        return channel

    def run(self):
        start_time = self.send_data(ALIVE_MESSAGE)
        if self.init_sleep is not None:
            self.bus.sleep(self.init_sleep)  # workaround for reset on COM open under Windows
        try:
            if self.device_id is None:
                # generate unique messages so that it can be identified
                # among received messages. The device writes all transmitted
                # and received messages. In case of transmitted messages there
                # is my DeviceID.
                unique_cmd = ALIVE_MESSAGE + b'-%d' % start_time.microseconds
                self.send_data(unique_cmd)
                print('Running self-detection... waiting for', unique_cmd)
                while len([packet for packet in self.recent_packets if unique_cmd in packet]) == 0:
                    self.update()
                ref = [packet for packet in self.recent_packets if unique_cmd in packet][0]
                print('Self-identified:', ref)
                addr, data = parse_lora_packet(ref)
                assert addr is not None, ref
                assert len(addr) == 1, ref
                self.device_id = addr[0]
            print('Using device_id', self.device_id)

            while True:
                self.update()
        except BusShutdownException:
            pass

    def draw(self):
        """
        Debug Draw
        """
        draw_lora_positions(self.debug_robot_poses)


# vim: expandtab sw=4 ts=4
